//############################################################################
//
// EZFB ~ Linux Frame Buffer API ~
//
// by James Lehman
// james@akrobiz.com
//
// began: Feb. 2001
//
// EZFB is intended to give C programmers an easy-to-use library of functions
// for drawing points, lines and strings of characters into the Linux frame
// buffer. It also provides a means of displaying bitmap images on any portion
// of the screen and capturing the contents of any portion of the screen to
// a standard bitmap graphics file. All of this is embodied in a set of
// functions that are polymorphic with regard to the characteristics of the
// frame buffer and the bitmap files used. In other words, it makes no
// difference what the screen height or width is or what the color depth is,
// all of the function calls are the same and will have appropriate and
// predictable effects.
//
// This software is distributed in the hope that it will be useful, but
// without any warranty; without even the implied warranty of merchantability
// or fitness for a particular purpose. See the GNU General Public License
// for more details.
//
//############################################################################
#include "ezfb_tty.h"
#include "ezfb_colors.h"
#include "ezfb_device.h"
#include "ezfb_pixels.h"

//############################################################################
int    ezfb_open_fb                 (struct ezfb* fb, char* dev_name);
int    ezfb_get_fb_var              (struct ezfb* fb);
int    ezfb_set_fb_var              (struct ezfb* fb);
int    ezfb_get_fb_fix              (struct ezfb* fb);
int    ezfb_screen_mem_size         (struct ezfb* fb);
int    ezfb_get_fb_mem              (struct ezfb* fb);
int    ezfb_make_mem_matrix         (struct ezfb* fb);
int    ezfb_make_screen_matrix      (struct ezfb* fb);
int    ezfb_make_off_screen_matrix  (struct ezfb* fb);
int    ezfb_find_screen             (struct ezfb* fb);
int    ezfb_off_find_screen         (struct ezfb* fb);
int    ezfb_save_mem                (struct ezfb* fb);
int    ezfb_init_base               (struct ezfb* fb, char* dev_name, int what_to_save);
int    ezfb_restore_mem             (struct ezfb* fb);
int    ezfb_restore_screen          (struct ezfb* fb);
void   ezfb_set_functions           (struct ezfb* fb);
void   ezfb_set_sig_traps           (               );

//############################################################################
struct ezfb* valid_ezfb[EZFB_MAX_FB_NUM + 1] = {NULL};
u_int        num_valid_ezfb             = 0;

//############################################################################
int ezfb_dev_exists(char* dev_name)
{
    int fd;
EZFB_FUNCTION_CALL_1
    if(0 > (fd = open(dev_name, O_RDWR)))
    {
#if EZFB_API_ERRORS
        perror("ezfb ERROR: FRAME BUFFER DOES NOT EXIST ");
        printf("(%s) FAILED", dev_name);
#endif
        ret = 0;
    }
    else
    {
#if EZFB_API_MESSAGES
        printf("FRAME BUFFER %s EXSISTS\n", dev_name);
        fflush(stdout);
#endif
        close(fd);
    }
EZFB_FUNCTION_RETURN(ret)
}

//############################################################################
int ezfb_num_exists(int num)
{
    char dev_name[16] = {0};
EZFB_FUNCTION_CALL_1
    strcpy(dev_name, EZFB_DEV_NAME);
    sprintf(&(dev_name[strlen(dev_name)]), "%d", num);
    ret = ezfb_dev_exists(dev_name);
EZFB_FUNCTION_RETURN(ret)
}

//############################################################################
u_int ezfb_name_to_num(char* dev_name) // returns last one or two digits as u_int
{
    if(isdigit(dev_name[strlen(dev_name) - 2]))
        return (u_int)atoi(&dev_name[strlen(dev_name) - 2]);
    return (u_int)atoi(&dev_name[strlen(dev_name) - 1]);
}

//############################################################################
int ezfb_open_fb(struct ezfb* fb, char* dev_name)
{
EZFB_FUNCTION_CALL_1
    if(0 > (fb->fd = open(dev_name, O_RDWR)))
    {
#if EZFB_API_ERRORS
        perror("ezfb ERROR: OPENING FILE DESCRIPTOR FAILED ");
        printf("(%s)\n", fb->dev_name);
#endif
        ret = 0;
    }
    else
    {
        strcpy(fb->dev_name, dev_name);
        fb->dev_num = ezfb_name_to_num(fb->dev_name);
#if EZFB_API_MESSAGES
        printf("FRAME BUFFER %s SUCCESSFULLY OPENED\n", fb->dev_name);
        fflush(stdout);
#endif
    }
EZFB_FUNCTION_RETURN(ret)
}

//############################################################################
int ezfb_get_fb_var(struct ezfb* fb)
{
EZFB_FUNCTION_CALL_0
    if(0 > ioctl(fb->fd, FBIOGET_VSCREENINFO, &fb->Var))
    {
#if EZFB_API_ERRORS
        perror("ezfb ERROR: FBIOGET_VSCREENINFO failed ");
        printf("(%s)\n", fb->dev_name);
#endif
    }
    else // ioctl did not fail
    {
        if(!fb->exists)
        {
            fb->bpp_was          = fb->Var.bits_per_pixel ;
            fb->xres_was         = fb->Var.xres           ;
            fb->yres_was         = fb->Var.yres           ;
            fb->xres_virtual_was = fb->Var.xres_virtual   ;
            fb->yres_virtual_was = fb->Var.yres_virtual   ;
            fb->xoffset_was      = fb->Var.xoffset        ;
            fb->yoffset_was      = fb->Var.yoffset        ;
        }
        fb->max_x = fb->Var.xres - 1;
        fb->max_y = fb->Var.yres - 1;
        fb->mid_x = fb->max_x / 2;
        fb->mid_y = fb->max_y / 2;
        ret       = 1;
    }
EZFB_FUNCTION_RETURN(ret)
}

//############################################################################
int ezfb_set_fb_var(struct ezfb* fb)
{
EZFB_FUNCTION_CALL_1
    if(0 > ioctl(fb->fd, FBIOPUT_VSCREENINFO, &fb->Var))
    {
#if EZFB_API_ERRORS
        perror("ezfb ERROR: FBIOPUT_VSCREENINFO failed ");
        printf("(%s)\n", fb->dev_name);
#endif
        ret = 0;
    }
EZFB_FUNCTION_RETURN(ret)
}

//############################################################################
int ezfb_get_fb_fix(struct ezfb* fb)
{
EZFB_FUNCTION_CALL_1
    if(0 > ioctl(fb->fd, FBIOGET_FSCREENINFO, &fb->Fix))
    {
#if EZFB_API_ERRORS
        perror("ezfb ERROR: FBIOGET_FSCREENINFO failed ");
        printf("(%s)\n", fb->dev_name);
#endif
        ret = 0;
    }    
EZFB_FUNCTION_RETURN(ret)
}

//############################################################################
int ezfb_screen_mem_size(struct ezfb* fb)
{
    switch(fb->Var.bits_per_pixel)
    {
        case  1: return fb->Var.yres_virtual * fb->Var.xres_virtual * sizeof(u_char) / 8;
        //--------------------------------------------------------------------
        case  2: return fb->Var.yres_virtual * fb->Var.xres_virtual * sizeof(u_char) / 4;
        //--------------------------------------------------------------------
        case  4: return fb->Var.yres_virtual * fb->Var.xres_virtual * sizeof(u_char) / 2;
        //--------------------------------------------------------------------
        case  8: return fb->Var.yres_virtual * fb->Var.xres_virtual * sizeof(u_char);
        //--------------------------------------------------------------------
        case 15: 
        case 16: return fb->Var.yres_virtual * fb->Var.xres_virtual * sizeof(u_short);
        //--------------------------------------------------------------------
        case 24: return fb->Var.yres_virtual * fb->Var.xres_virtual * sizeof(u_char) * 3;
        //--------------------------------------------------------------------
        case 32: return fb->Var.yres_virtual * fb->Var.xres_virtual * sizeof(u_int);
        //--------------------------------------------------------------------
    }
    return 0;
}    
//############################################################################
int ezfb_get_fb_mem(struct ezfb* fb)
{
EZFB_FUNCTION_CALL_1

#if EZFB_NO_MMAP
    fb->mem_uchar = (u_char*)fb->Fix.smem_start;
#else
    fb->mem_uchar = (u_char*)mmap(   NULL
                                   , fb->Fix.smem_len
                                   , (PROT_READ | PROT_WRITE)
                                   , MAP_SHARED
                                   , fb->fd
                                   , 0
                                 );
#endif
    if(fb->mem_uchar == MAP_FAILED)
    {
#if EZFB_API_ERRORS
        perror("ezfb ERROR: MEMORY MAP FOR FRAME BUFFER DEVICE FAILED ");
        printf("(%s)\n", fb->dev_name);
#endif
        ret = 0;
    }
    else                
    {
        fb->mem_last_uchar = fb->mem_uchar + fb->Fix.smem_len;
        fb->mem_ushort     = (u_short*)fb->mem_uchar;
        fb->mem_uint       = (u_int*  )fb->mem_uchar;
    }
EZFB_FUNCTION_RETURN(ret)
}

//############################################################################
int ezfb_make_mem_matrix(struct ezfb* fb)
{   
    u_int y;
EZFB_FUNCTION_CALL
    if(fb->exists)
    { 
        if(fb->mem_row_uchar )    free(fb->mem_row_uchar );
        if(fb->mem_row_uint  )    free(fb->mem_row_uint  );
        if(fb->mem_row_ushort)    free(fb->mem_row_ushort);

        fb->mem_row_uchar  = NULL;
        fb->mem_row_uint   = NULL;
        fb->mem_row_ushort = NULL;

        switch(fb->Var.bits_per_pixel)
        {
            case  1: fb->mem_row_uchar = (u_char**)
                         malloc(fb->Var.yres_virtual * sizeof(u_char*));

                     for(y = 0; y < fb->Var.yres_virtual; y++)
                        fb->mem_row_uchar[y] = &(fb->mem_uchar[y * fb->Var.xres_virtual / 8]);
                     break;
            //----------------------------------------------------------------
            case  2: fb->mem_row_uchar = (u_char**)
                         malloc(fb->Var.yres_virtual * sizeof(u_char*));

                     for(y = 0; y < fb->Var.yres_virtual; y++)
                        fb->mem_row_uchar[y] = &(fb->mem_uchar[y * fb->Var.xres_virtual / 4]);
                     break;
            //----------------------------------------------------------------
            case  4: fb->mem_row_uchar = (u_char**)
                         malloc(fb->Var.yres_virtual * sizeof(u_char*));

                     for(y = 0; y < fb->Var.yres_virtual; y++)
                        fb->mem_row_uchar[y] = &(fb->mem_uchar[y * fb->Var.xres_virtual / 2]);
                     break;
            //----------------------------------------------------------------
            case  8: fb->mem_row_uchar = (u_char**)
                         malloc(fb->Var.yres_virtual * sizeof(u_char*));

                     for(y = 0; y < fb->Var.yres_virtual; y++)
                        fb->mem_row_uchar[y] = &(fb->mem_uchar[y * fb->Var.xres_virtual]);
                     break;
            //----------------------------------------------------------------
            case 15: 
            case 16: fb->mem_row_ushort = (u_short**)
                         malloc(fb->Var.yres_virtual * sizeof(u_short*));

                     for(y = 0; y < fb->Var.yres_virtual; y++)
                        fb->mem_row_ushort[y] = &(fb->mem_ushort[y * fb->Var.xres_virtual]);
                     break;
            //----------------------------------------------------------------
            case 24: fb->mem_row_uchar = (u_char**)
                         malloc(fb->Var.yres_virtual * sizeof(u_char*));

                     for(y = 0; y < fb->Var.yres_virtual; y++)
                        fb->mem_row_uchar[y] = &(fb->mem_uchar[y * fb->Var.xres_virtual * 3]);
                     break;
            //----------------------------------------------------------------
            case 32: fb->mem_row_uint = (u_int** )
                         malloc(fb->Var.yres_virtual * sizeof(u_int*));

                     for(y = 0; y < fb->Var.yres_virtual; y++)
                        fb->mem_row_uint[y] = &(fb->mem_uint[y * fb->Var.xres_virtual]);
                     break;
            //----------------------------------------------------------------
        }
#if EZFB_API_MESSAGES
        printf("display matrix made and set\n");
        fflush(stdout);
#endif
    }
EZFB_FUNCTION_RETURN(fb->exists)
}

//############################################################################
int ezfb_make_screen_matrix(struct ezfb* fb)
{   
EZFB_FUNCTION_CALL
    if(fb->exists)
    { 
        if(fb->mem_row_uchar_xoff )    free(fb->mem_row_uchar_xoff );
        if(fb->mem_row_ushort_xoff)    free(fb->mem_row_ushort_xoff);
        if(fb->mem_row_uint_xoff  )    free(fb->mem_row_uint_xoff  );

        fb->mem_row_uchar_xoff  = NULL;
        fb->mem_row_ushort_xoff = NULL;
        fb->mem_row_uint_xoff   = NULL;

        fb->screen_row_uchar    = NULL;
        fb->screen_row_ushort   = NULL;
        fb->screen_row_uint     = NULL;

        switch(fb->Var.bits_per_pixel)
        {
            case  1: 
            case  2: 
            case  4: 
            case  8: 
            case 24: fb->mem_row_uchar_xoff = (u_char**)
                         malloc(fb->Var.yres_virtual * sizeof(u_char*));
                     break;
            //----------------------------------------------------------------
            case 15: 
            case 16: fb->mem_row_ushort_xoff = (u_short**)
                         malloc(fb->Var.yres_virtual * sizeof(u_short*));
                     break;
            //----------------------------------------------------------------
            case 32: fb->mem_row_uint_xoff = (u_int** )
                         malloc(fb->Var.yres_virtual * sizeof(u_int*));
                     break;
            //----------------------------------------------------------------
        }
#if EZFB_API_MESSAGES
        printf("screen matrix made.\n");
        fflush(stdout);
#endif
    }
EZFB_FUNCTION_RETURN(fb->exists)
}

//############################################################################
int ezfb_make_off_screen_matrix(struct ezfb* fb)
{   
EZFB_FUNCTION_CALL
    if(fb->exists)
    { 
        fb->off_screen_row_uchar  = NULL;
        fb->off_screen_row_ushort = NULL;
        fb->off_screen_row_uint   = NULL;

#if EZFB_API_MESSAGES
        printf("off screen matrix made.\n");
        fflush(stdout);
#endif
    }
EZFB_FUNCTION_RETURN(fb->exists)
}

//############################################################################
int ezfb_find_screen(struct ezfb* fb)
{   
    u_int y;
EZFB_FUNCTION_CALL
    if(fb->exists)
    { 
        switch(fb->Var.bits_per_pixel)
        {
            case  1: for(y = 0; y < fb->Var.yres_virtual; y++)
                         fb->mem_row_uchar_xoff[y] = &(fb->mem_row_uchar[y][fb->Var.xoffset / 8]);
                     fb->screen_row_uchar = &(fb->mem_row_uchar_xoff[fb->Var.yoffset]);
                     break;
            //----------------------------------------------------------------
            case  2: for(y = 0; y < fb->Var.yres_virtual; y++)
                         fb->mem_row_uchar_xoff[y] = &(fb->mem_row_uchar[y][fb->Var.xoffset / 4]);
                     fb->screen_row_uchar = &(fb->mem_row_uchar_xoff[fb->Var.yoffset]);
                     break;
            //----------------------------------------------------------------
            case  4: for(y = 0; y < fb->Var.yres_virtual; y++)
                         fb->mem_row_uchar_xoff[y] = &(fb->mem_row_uchar[y][fb->Var.xoffset / 2]);
                     fb->screen_row_uchar = &(fb->mem_row_uchar_xoff[fb->Var.yoffset]);
                     break;
            //----------------------------------------------------------------
            case  8: for(y = 0; y < fb->Var.yres_virtual; y++)
                         fb->mem_row_uchar_xoff[y] = &(fb->mem_row_uchar[y][fb->Var.xoffset]);
                     fb->screen_row_uchar = &(fb->mem_row_uchar_xoff[fb->Var.yoffset]);
                     break;
            //----------------------------------------------------------------
            case 15: 
            case 16: for(y = 0; y < fb->Var.yres_virtual; y++)
                         fb->mem_row_ushort_xoff[y] = &(fb->mem_row_ushort[y][fb->Var.xoffset]);
                     fb->screen_row_ushort = &(fb->mem_row_ushort_xoff[fb->Var.yoffset]);
                     break;
            //----------------------------------------------------------------
            case 24: for(y = 0; y < fb->Var.yres_virtual; y++)
                         fb->mem_row_uchar_xoff[y] = &(fb->mem_row_uchar[y][fb->Var.xoffset * 3]);
                     fb->screen_row_uchar = &(fb->mem_row_uchar_xoff[fb->Var.yoffset]);
                     break;
            //----------------------------------------------------------------
            case 32: for(y = 0; y < fb->Var.yres_virtual; y++)
                         fb->mem_row_uint_xoff[y] = &(fb->mem_row_uint[y][fb->Var.xoffset]);
                     fb->screen_row_uint = &(fb->mem_row_uint_xoff[fb->Var.yoffset]);
                     break;
            //----------------------------------------------------------------
        }
#if EZFB_API_MESSAGES
        printf("screen found\n");
        fflush(stdout);
#endif
    }
EZFB_FUNCTION_RETURN(fb->exists)
}

//############################################################################
int ezfb_find_off_screen(struct ezfb* fb)
{   
    u_int y_off;
EZFB_FUNCTION_CALL           
    if(fb->exists)
    { 
        y_off =   (fb->Var.yres_virtual >= (fb->Var.yres * 2))
                ? (fb->Var.yres)
                : (0);

        switch(fb->Var.bits_per_pixel)
        {
            case  1: 
            case  2: 
            case  4: 
            case  8:
            case 24: fb->off_screen_row_uchar = &(fb->mem_row_uchar_xoff[fb->Var.yoffset + y_off]);
                     break;
            //----------------------------------------------------------------
            case 15: 
            case 16: fb->off_screen_row_ushort = &(fb->mem_row_ushort_xoff[fb->Var.yoffset + y_off]);
                     break;
            //----------------------------------------------------------------
            case 32: fb->off_screen_row_uint = &(fb->mem_row_uint_xoff[fb->Var.yoffset + y_off]);
                     break;
            //----------------------------------------------------------------
        }
#if EZFB_API_MESSAGES
            printf("off screen found\n");
            fflush(stdout);
#endif
    }
EZFB_FUNCTION_RETURN(fb->exists)
}

//############################################################################
int ezfb_save_mem(struct ezfb* fb)
{
EZFB_FUNCTION_CALL
    if(fb->exists)
    {
        switch(fb->what_was_saved)
        {
            //----------------------------------------------------------------
            case EZFB_SAVE_MEM:      fb->data_was = (u_char*)malloc(fb->Fix.smem_len * sizeof(u_char));
#if EZFB_API_MESSAGES
                                    printf("SAVING CONTENTS OF VIDEO RAM ... {%d bytes}\na lot of video memory takes a while.\n", (fb->Fix.smem_len * sizeof(u_char)));
                                    fflush(stdout);
#endif
                                    memcpy((void*)fb->data_was, (const void*)fb->mem_uchar, fb->Fix.smem_len); // this can take a while!
#if EZFB_API_MESSAGES
                                    printf("CONTENTS OF VIDEO RAM SAVED\n");
                                    fflush(stdout);
#endif
                                    break;
            //----------------------------------------------------------------
            case EZFB_SAVE_SCREEN:  fb->data_was = (u_char*)malloc(ezfb_screen_mem_size(fb));
#if EZFB_API_MESSAGES
                                    printf("SAVING CONTENTS OF SCREEN ... {%d bytes}\n", ezfb_screen_mem_size(fb));                                                                
                                    fflush(stdout);
#endif
                                    switch(fb->Var.bits_per_pixel)
                                    {
                                        case  1: 
                                        case  2: 
                                        case  4: 
                                        case  8:
                                        case 24: memcpy((void*)fb->data_was, (const void*)&(fb->screen_row_uchar[0][0]), ezfb_screen_mem_size(fb));
                                                 break;
                                        //----------------------------------------------------------------
                                        case 15: 
                                        case 16: memcpy((void*)fb->data_was, (const void*)&(fb->screen_row_ushort[0][0]), ezfb_screen_mem_size(fb));
                                                 break;
                                        //----------------------------------------------------------------
                                        case 32: memcpy((void*)fb->data_was, (const void*)&(fb->screen_row_uint[0][0]), ezfb_screen_mem_size(fb));
                                                 break;
                                        //----------------------------------------------------------------
                                    }
#if EZFB_API_MESSAGES
                                    printf("CONTENTS OF SCREEN SAVED\n");
                                    fflush(stdout);
#endif
                                    break;
            //----------------------------------------------------------------
            case EZFB_NO_SAVE:        fb->data_was = NULL; // just to be sure
#if EZFB_API_MESSAGES
                                    printf("CONTENTS OF VIDEO RAM NOT SAVED\n");
                                    fflush(stdout);
#endif
                                    break;
            //----------------------------------------------------------------
        }
    }
EZFB_FUNCTION_RETURN(fb->exists)
}

//############################################################################
// ezfb constructor
// An ezfb object should be instantiated empty like this:
// struct ezfb fb = {0};
//
// Macros for what_to_save
// EZFB_SAVE_MEM        0
// EZFB_SAVE_SCREEN     1
// EZFB_NO_SAVE         2
//----------------------------------------------------------------------------
int ezfb_init_base(struct ezfb* fb, char* dev_name, int what_to_save)
{
EZFB_FUNCTION_CALL_1
    srandom(time(NULL)); // set random seed to time now
    ezfb_set_sig_traps();
    if(fb)
    {
        if(valid_ezfb[ezfb_name_to_num(dev_name)] == NULL)
        {
            if(!fb->exists)
            {   
#if !EZFB_NO_TTY0
                if(ret && fb->dev_num == 0) // /dev/fb0 console
                    ret = ezfb_tty_graphics();
#endif
                if(ret)    ret = ezfb_open_fb    (fb, dev_name); // sets fb->dev_name, fb->dev_num
                if(ret)    ret = ezfb_get_fb_fix (fb);
                if(ret)    ret = ezfb_get_fb_var (fb);

                ezfb_set_functions(fb);
                fb->what_was_saved = what_to_save;

                if(ret)    ret = ezfb_get_fb_mem (fb);
                if(ret)    ret = ezfb_init_cmap  (fb);
                if(ret)
                {
                    fb->exists = 1;

                    fb->mem_row_uchar       = NULL;
                    fb->mem_row_ushort      = NULL;
                    fb->mem_row_uint        = NULL;

                    fb->mem_row_uchar_xoff  = NULL;
                    fb->mem_row_ushort_xoff = NULL;
                    fb->mem_row_uint_xoff   = NULL;

                    fb->screen_row_uchar    = NULL;
                    fb->screen_row_ushort   = NULL;
                    fb->screen_row_uint     = NULL;

                    ret &= ezfb_make_mem_matrix       (fb);
                    ret &= ezfb_make_screen_matrix    (fb);
                    ret &= ezfb_find_screen           (fb);
                    ret &= ezfb_save_mem              (fb);
                    ret &= ezfb_find_black_gray_white (fb);

                    valid_ezfb[fb->dev_num] = fb;
                    num_valid_ezfb++;
#if EZFB_API_MESSAGES
                    printf("ezfb %s SUCCESSFULLY INITIALIZED\n", fb->dev_name);
                    fflush(stdout);
#endif
                }
            } // if(!fb->exists)
            else
            {
#if EZFB_API_ERRORS
                fprintf(stderr, "ezfb ERROR: ezfb_init_base( %s ) CALLED MORE THAN ONCE!\n", fb->dev_name);
                fflush(stderr);
#endif
                ret = 0;
            }
        } // if(valid_ezfb[ezfb_name_to_num(dev_name)] == NULL)
        else
        {
#if EZFB_API_ERRORS
            fprintf(stderr, "ezfb ERROR: FRAME BUFFER %s ALREADY OPEN BY THIS APPLICATION\n", dev_name);
            fflush(stderr);
#endif
            ret = 0;
        }
    } // if(fb)
    else
    {
#if EZFB_API_ERRORS
        fprintf(stderr, "ezfb ERROR: CAN NOT INITIALIZE NULL POINTER\n");
        fflush(stderr);
#endif
        ret = 0;
    }
EZFB_FUNCTION_RETURN(ret)
}

//############################################################################
int ezfb_init(struct ezfb* fb, int what_to_save) // each call attmepts to init next fb
{
EZFB_FUNCTION_CALL_1
    if(num_valid_ezfb <= EZFB_MAX_FB_NUM)
        ret = ezfb_init_num(fb, num_valid_ezfb, what_to_save);
#if EZFB_API_ERRORS
    if(!ret)
    {
        fprintf(stderr, "ezfb ERROR: NO MORE VALID FRAME BUFFER DEVICES\n");
        fflush(stderr);
    }
#endif
EZFB_FUNCTION_RETURN(ret)
}

//############################################################################
int ezfb_init_num(struct ezfb* fb, u_int dev_num, int what_to_save) // from 0 to 31
{
    char dev_name[EZFB_MAX_NAME_LENGTH] = {'\0'};
EZFB_FUNCTION_CALL_1
    if(dev_num >= 0 && dev_num <= EZFB_MAX_FB_NUM)
    {
        sprintf(dev_name, "%s%u", EZFB_DEV_NAME, dev_num);
        if(!(ret = ezfb_init_base(fb, dev_name, what_to_save)))
        {
            sprintf(dev_name, "%s/%d", EZFB_DEV_NAME, dev_num);
            ret = ezfb_init_base(fb, dev_name, what_to_save);
        }
    }
    else
    {
#if EZFB_API_ERRORS
        fprintf(stderr, "ezfb ERROR: %d OUT OF RANGE!\n", dev_num);
        fflush(stderr);
#endif
        ret = 0;
    }
EZFB_FUNCTION_RETURN(ret)
}

//############################################################################
int ezfb_init_named(struct ezfb* fb, char* dev_name, int what_to_save) // dev_name is like: "/dev/fbN"
{
EZFB_FUNCTION_CALL_1
    if(!ezfb_init_base(fb, dev_name, what_to_save))
        ret = 0;
EZFB_FUNCTION_RETURN(ret)
}

//############################################################################
int ezfb_set_bpp(struct ezfb* fb, u_int bpp_wanted)
{
EZFB_FUNCTION_CALL_0
    if(fb->exists)
    {
        if(fb->Var.bits_per_pixel == bpp_wanted)
        {
#if EZFB_API_NIT_PICKING
            fprintf(stderr, "ezfb ERROR: ALREADY AT %d bpp.\n", bpp_wanted);
            fflush(stderr);
#endif
            ret = 1; // error, but return ok
        }
        else
        {
            fb->Var.bits_per_pixel = bpp_wanted;
            ret  = ezfb_set_fb_var (fb);
            ret &= ezfb_get_fb_var (fb);            
            if(fb->Var.bits_per_pixel != bpp_wanted) // bpp not supported
            {
#if EZFB_API_ERRORS
                fprintf(stderr, "ezfb ERROR: ezfb_set_bpp(fb, %d) FAILED\n", bpp_wanted);
                fflush(stderr);
#endif
                ret = 0;
            }
            else
            {
                ezfb_set_functions            (fb);
                ret &= ezfb_set_cmap          (fb);
                ret &= ezfb_get_cmap          (fb);
                ret &= ezfb_get_sizeof_colors (fb);

                if(fb->Var.bits_per_pixel <= 8)
                {
                    fb->colors.fbcmap.len = 0x01 << fb->Var.bits_per_pixel;
                    ret &= ezfb_set_cmap_utility(fb);
                }
                else if(    fb->Fix.visual == FB_VISUAL_DIRECTCOLOR
                         || fb->Fix.visual == FB_VISUAL_PSEUDOCOLOR                                            
                       )
                {
                    fb->colors.fbcmap.len = 256;
                    ret &= ezfb_set_cmap_directcolor(fb);
                }
                else
                    fb->colors.fbcmap.len = 0;

                ret &= ezfb_find_black_gray_white (fb);
                ret &= ezfb_make_mem_matrix       (fb);
                ret &= ezfb_make_screen_matrix    (fb);
                ret &= ezfb_find_screen           (fb);
                fb->bpp_was_changed = 1;
            }
        }
    }
#if EZFB_API_ERRORS
    else
    {
        fprintf(stderr, "ezfb ERROR: (ezfb_set_bpp) FB NOT INITIALIZED\n");
        fflush(stderr);
    }
#endif
EZFB_FUNCTION_RETURN(ret)
}

//############################################################################
int ezfb_set_xy_virtual(struct ezfb* fb, u_int xvirt, u_int yvirt)
{
EZFB_FUNCTION_CALL_0
    if(fb->exists)
    {
        fb->Var.xres_virtual = xvirt;
        fb->Var.yres_virtual = yvirt;
        ret  = ezfb_set_fb_var    (fb);
        ret &= ezfb_get_fb_var    (fb);
        //--------------------------------------------------------------------
        if(    (fb->Var.xres_virtual != xvirt)
            || (fb->Var.yres_virtual != yvirt)
          ) // we did not get what we asked for
        {
#if EZFB_API_ERRORS
            fprintf( stderr,
                     "ezfb ERROR: ezfb_set_xy_virtual(fb, %d, %d) FAILED\n",
                     xvirt, yvirt
                   );
            fflush(stderr);
#endif
            //----------------------------------------------------------------
            if(    (fb->Var.xres_virtual != fb->xres_virtual_was)
                || (fb->Var.yres_virtual != fb->yres_virtual_was)
              ) // it was changed
            {
#if EZFB_API_MESSAGES
                printf("ezfb %s xres_virtual : %d, yres_virtual : %d\n",
                       fb->dev_name, fb->Var.xres_virtual, fb->Var.yres_virtual);
                fflush(stdout);
                fb->resolution_changed = 1;
                ret &= ezfb_make_mem_matrix    (fb);
                ret &= ezfb_make_screen_matrix (fb);
                ret &= ezfb_find_screen        (fb);
#endif
            }
            ret = 0;
            //----------------------------------------------------------------
        }
        //--------------------------------------------------------------------
        else if(    (fb->Var.xres_virtual != fb->xres_virtual_was)
                 || (fb->Var.yres_virtual != fb->yres_virtual_was)
               ) // it was changed to what we wanted
        {
#if EZFB_API_MESSAGES
            printf("ezfb %s xres_virtual : %d, yres_virtual : %d\n",
                   fb->dev_name, fb->Var.xres_virtual, fb->Var.yres_virtual);
            fflush(stdout);
#endif
            fb->resolution_changed = 1;
            ret &= ezfb_make_mem_matrix    (fb);
            ret &= ezfb_make_screen_matrix (fb);
            ret &= ezfb_find_screen        (fb);
        }
#if EZFB_API_MESSAGES
        //--------------------------------------------------------------------
        else // it was not changed
        {
            printf("ezfb %s xres_virtual : %d, yres_virtual : %d : no change\n",
                   fb->dev_name, fb->Var.xres_virtual, fb->Var.yres_virtual);
            fflush(stdout);
        }
        //--------------------------------------------------------------------
#endif
    }
#if EZFB_API_ERRORS
    else
    {
        fprintf(stderr, "ezfb ERROR: (ezfb_set_xy_virtual) FB NOT INITIALIZED\n");
        fflush(stderr);
    }
#endif
EZFB_FUNCTION_RETURN(ret)
}

//############################################################################
int ezfb_set_xy_offsets(struct ezfb* fb, int xoff, int yoff)
{
EZFB_FUNCTION_CALL_0
    if(fb->exists)
    {
        if(    (xoff >= 0)
            && (xoff <= (int)(fb->Var.xres_virtual - fb->Var.xres))
          )
            fb->Var.xoffset = xoff;
        //--------------------------------------------------------------------
        if(    (yoff >= 0)
            && (yoff <= (int)(fb->Var.yres_virtual - fb->Var.yres))
          )
            fb->Var.yoffset = yoff;
        //--------------------------------------------------------------------
        ret  = ezfb_set_fb_var    (fb);
        ret &= ezfb_get_fb_var    (fb);
        //--------------------------------------------------------------------
        if(    ((int)fb->Var.xoffset != xoff)
            || ((int)fb->Var.yoffset != yoff) // we didn't get what we wanted
          )
        {
#if EZFB_API_NIT_PICKING
            fprintf(stderr, "ezfb ERROR: ezfb_set_xy_offsets(fb, %d, %d) FAILED\n", xoff, yoff);
            fflush(stderr);
#endif            
        }
        //--------------------------------------------------------------------
        if(    (fb->Var.xoffset != fb->xoffset_was)
            || (fb->Var.yoffset != fb->yoffset_was) // we changed something
          )
        {
#if EZFB_API_NIT_PICKING
            printf("ezfb %s xoffset : %d, yoffset :  %d\n", fb->dev_name, fb->Var.xoffset, fb->Var.yoffset);
            fflush(stdout);
#endif
            fb->resolution_changed = 1;
            //ezfb_find_screen(fb);
        }
        //--------------------------------------------------------------------
#if EZFB_API_NIT_PICKING
        else // we didn't change anything
        {
            printf("ezfb %s xoffset : %d, yoffset :  %d : no change\n", fb->dev_name, fb->Var.xoffset, fb->Var.yoffset);
            fflush(stdout);
            ret = 0;
        }
#endif
    }
    //------------------------------------------------------------------------
#if EZFB_API_ERRORS
    else
    {
        fprintf(stderr, "ezfb ERROR: (ezfb_set_xy_offsets) FB NOT INITIALIZED\n");
        fflush(stderr);
    }
#endif
EZFB_FUNCTION_RETURN(ret)
}

//############################################################################
//int ezfb_set_resolution(struct ezfb* fb, int new_xres, int new_yres)
//{       //TODO: make this work!
//EZFB_FUNCTION_CALL_1
//    if(fb->exists)
//    {
//        fb->Var.xres = new_xres;
//        fb->Var.yres = new_yres;
//        ret &= ezfb_set_fb_var(fb);
//        ret &= ezfb_get_fb_var(fb);
//        if(ret) fb->resolution_changed = 1;
//    }
//    else
//    {
//#if EZFB_API_ERRORS
//        fprintf(stderr, "ezfb ERROR: (ezfb_set_resolution) FB NOT INITIALIZED");
//        fflush(stderr);
//#endif
//        ret = 0;
//    }
//EZFB_FUNCTION_RETURN(ret)
//}

//############################################################################
int ezfb_restore_mem(struct ezfb* fb)
{
EZFB_FUNCTION_CALL_1
    if(fb->exists && fb->data_was)
    {
        memcpy((void*)fb->mem_uchar, (const void*)fb->data_was, fb->Fix.smem_len);
        free(fb->data_was);
        fb->data_was = NULL;
#if EZFB_API_MESSAGES
        printf("ezfb %s memory contents restored.\n", fb->dev_name);
        fflush(stdout);
#endif
    }
    else
    {
#if EZFB_API_ERRORS
        fprintf(stderr, "ezfb ERROR: fb->data_was NEVER SAVED!\n");
        fflush(stderr);
#endif
        ret = 0;
    }
EZFB_FUNCTION_RETURN(ret)
}

//############################################################################
int ezfb_restore_screen(struct ezfb* fb)
{
EZFB_FUNCTION_CALL_1
    if(fb->exists && fb->data_was)
    {
        switch(fb->Var.bits_per_pixel)
        {
            case  1: 
            case  2: 
            case  4: 
            case  8:
            case 24: memcpy((void*)&(fb->screen_row_uchar[0][0]), (const void*)fb->data_was, ezfb_screen_mem_size(fb));
                     break;
            //----------------------------------------------------------------
            case 15: 
            case 16: memcpy((void*)&(fb->screen_row_ushort[0][0]), (const void*)fb->data_was, ezfb_screen_mem_size(fb));
                     break;
            //----------------------------------------------------------------
            case 32: memcpy((void*)&(fb->screen_row_uint[0][0]), (const void*)fb->data_was, ezfb_screen_mem_size(fb));
                     break;
            //----------------------------------------------------------------
        }
        free(fb->data_was);
        fb->data_was = NULL;
#if EZFB_API_MESSAGES
        printf("ezfb %s screen contents restored.\n", fb->dev_name);
        fflush(stdout);
#endif
    }
    else
    {
#if EZFB_API_ERRORS
        fprintf(stderr, "ezfb ERROR: fb->data_was NEVER SAVED!\n");
        fflush(stderr);
#endif
        ret = 0;
    }
EZFB_FUNCTION_RETURN(ret)
}

//############################################################################
int ezfb_clear_mem(struct ezfb* fb)
{
EZFB_FUNCTION_CALL
    if(fb->exists)
        memset((void*)fb->mem_uchar, 0x00, fb->Fix.smem_len);
#if EZFB_API_ERRORS
    else
    {
        fprintf(stderr, "ezfb ERROR: (ezfb_clear_mem) FB NOT INITIALIZED\n");
        fflush(stderr);
    }
#endif
EZFB_FUNCTION_RETURN(fb->exists)
}

//############################################################################
int ezfb_clear_screen_at_offsets(struct ezfb* fb, int xoff, int yoff)
{
    u_int     y;
    u_char*   screen;
EZFB_FUNCTION_CALL
    if(fb->exists)
    {
        switch(fb->Var.bits_per_pixel)
        {
            case  1: screen = fb->mem_uchar + (yoff * fb->Var.xres_virtual + xoff) / 8;
                     for(y = 0; y < fb->Var.yres; y++)
                         memset(   (void*)(screen + y * fb->Var.xres_virtual / 8)
                                 , 0x00
                                 , fb->Var.xres / 8
                               );
                     break;
            //---------------------------------------------------------------- 
            case  2: screen = fb->mem_uchar + (yoff * fb->Var.xres_virtual + xoff) / 4;
                     for(y = 0; y < fb->Var.yres; y++)
                         memset(   (void*)(screen + y * fb->Var.xres_virtual / 4)
                                 , 0x00
                                 , fb->Var.xres / 4
                               );
                     break;
            //---------------------------------------------------------------- 
            case  4: screen = fb->mem_uchar + (yoff * fb->Var.xres_virtual + xoff) / 2;
                     for(y = 0; y < fb->Var.yres; y++)
                         memset(   (void*)(screen + y * fb->Var.xres_virtual / 2)
                                 , 0x00
                                 , fb->Var.xres / 2
                               );
                     break;
            //---------------------------------------------------------------- 
            case  8: screen = fb->mem_uchar + (yoff * fb->Var.xres_virtual + xoff);
                     for(y = 0; y < fb->Var.yres; y++)
                         memset(   (void*)(screen + y * fb->Var.xres_virtual)
                                 , 0x00
                                 , fb->Var.xres
                               );
                     break;
            //---------------------------------------------------------------- 
            case 15: 
            case 16: screen = fb->mem_uchar + (yoff * fb->Var.xres_virtual + xoff) * 2;
                     for(y = 0; y < fb->Var.yres; y++)
                         memset(   (void*)(screen + y * fb->Var.xres_virtual * 2)
                                 , 0x00
                                 , fb->Var.xres * 2
                               );
                     break;
            //---------------------------------------------------------------- 
            case 24: screen = fb->mem_uchar + (yoff * fb->Var.xres_virtual + xoff) * 3;
                     for(y = 0; y < fb->Var.yres; y++)
                         memset(   (void*)(screen + y * fb->Var.xres_virtual * 3)
                                 , 0x00
                                 , fb->Var.xres * 3
                               );
                     break;
            //---------------------------------------------------------------- 
            case 32: screen = fb->mem_uchar + (yoff * fb->Var.xres_virtual + xoff) * 4;
                     for(y = 0; y < fb->Var.yres; y++)
                         memset(   (void*)(screen + y * fb->Var.xres_virtual * 4)
                                 , 0x00
                                 , fb->Var.xres * 4
                               );
                     break;
            //---------------------------------------------------------------- 
            default: exit(333);
        }
    }
#if EZFB_API_ERRORS
    else
    {
        fprintf(stderr, "ezfb ERROR: (ezfb_clear_screen_at_offsets) FB NOT INITIALIZED\n");
        fflush(stderr);
    }
#endif
EZFB_FUNCTION_RETURN(fb->exists)
}

//############################################################################
int ezfb_clear_screen(struct ezfb* fb)
{
    u_int xx, y;
EZFB_FUNCTION_CALL
    if(fb->exists)
    {
        switch(fb->Var.bits_per_pixel)
        {
            case  1: xx = fb->Var.xres / 8;
                     for(y = 0; y < fb->Var.yres; y++)
                         memset(   (void*)(&(fb->screen_row_uchar[y][0]))
                                 , 0x00
                                 , xx
                               );
                     break;
            //---------------------------------------------------------------- 
            case  2: xx = fb->Var.xres / 4;
                     for(y = 0; y < fb->Var.yres; y++)
                         memset(   (void*)(&(fb->screen_row_uchar[y][0]))
                                 , 0x00
                                 , xx
                               );
                     break;
            //---------------------------------------------------------------- 
            case  4: xx = fb->Var.xres / 2;
                     for(y = 0; y < fb->Var.yres; y++)
                         memset(   (void*)(&(fb->screen_row_uchar[y][0]))
                                 , 0x00
                                 , xx
                               );
                     break;
            //---------------------------------------------------------------- 
            case  8: for(y = 0; y < fb->Var.yres; y++)
                         memset(   (void*)(&(fb->screen_row_uchar[y][0]))
                                 , 0x00
                                 , fb->Var.xres
                               );
                     break;
            //---------------------------------------------------------------- 
            case 15: 
            case 16: xx = fb->Var.xres * 2;
                     for(y = 0; y < fb->Var.yres; y++)
                         memset(   (void*)(&(fb->screen_row_ushort[y][0]))
                                 , 0x00
                                 , xx
                               );
                     break;
            //---------------------------------------------------------------- 
            case 24: xx = fb->Var.xres * 3;
                     for(y = 0; y < fb->Var.yres; y++)
                         memset(   (void*)(&(fb->screen_row_uchar[y][0]))
                                 , 0x00
                                 , xx
                               );
                     break;
            //---------------------------------------------------------------- 
            case 32: xx = fb->Var.xres * 4;
                     for(y = 0; y < fb->Var.yres; y++)
                         memset(   (void*)(&(fb->screen_row_uint[y][0]))
                                 , 0x00
                                 , xx
                               );
                     break;
            //---------------------------------------------------------------- 
            default: exit(333);
        }
    }
#if EZFB_API_ERRORS
    else
    {
        fprintf(stderr, "ezfb ERROR: (ezfb_clear_screen) FB NOT INITIALIZED\n");
        fflush(stderr);
    }
#endif
EZFB_FUNCTION_RETURN(fb->exists)
}

//############################################################################
int ezfb_clear_off_screen(struct ezfb* fb)
{
    u_int xx, y;
EZFB_FUNCTION_CALL
    if(fb->exists)
    {
        switch(fb->Var.bits_per_pixel)
        {
            case  1: xx = fb->Var.xres / 8;
                     for(y = 0; y < fb->Var.yres; y++)
                         memset(   (void*)(&(fb->off_screen_row_uchar[y][0]))
                                 , 0x00
                                 , xx
                               );
                     break;
            //---------------------------------------------------------------- 
            case  2: xx = fb->Var.xres / 4;
                     for(y = 0; y < fb->Var.yres; y++)
                         memset(   (void*)(&(fb->off_screen_row_uchar[y][0]))
                                 , 0x00
                                 , xx
                               );
                     break;
            //---------------------------------------------------------------- 
            case  4: xx = fb->Var.xres / 2;
                     for(y = 0; y < fb->Var.yres; y++)
                         memset(   (void*)(&(fb->off_screen_row_uchar[y][0]))
                                 , 0x00
                                 , xx
                               );
                     break;
            //---------------------------------------------------------------- 
            case  8: for(y = 0; y < fb->Var.yres; y++)
                         memset(   (void*)(&(fb->off_screen_row_uchar[y][0]))
                                 , 0x00
                                 , fb->Var.xres
                               );
                     break;
            //---------------------------------------------------------------- 
            case 15: 
            case 16: xx = fb->Var.xres * 2;
                     for(y = 0; y < fb->Var.yres; y++)
                         memset(   (void*)(&(fb->off_screen_row_ushort[y][0]))
                                 , 0x00
                                 , xx
                               );
                     break;
            //---------------------------------------------------------------- 
            case 24: xx = fb->Var.xres * 3;
                     for(y = 0; y < fb->Var.yres; y++)
                         memset(   (void*)(&(fb->off_screen_row_uchar[y][0]))
                                 , 0x00
                                 , xx
                               );
                     break;
            //---------------------------------------------------------------- 
            case 32: xx = fb->Var.xres * 4;
                     for(y = 0; y < fb->Var.yres; y++)
                         memset(   (void*)(&(fb->off_screen_row_uint[y][0]))
                                 , 0x00
                                 , xx
                               );
                     break;
            //---------------------------------------------------------------- 
            default: exit(333);
        }
    }
#if EZFB_API_ERRORS
    else
    {
        fprintf(stderr, "ezfb ERROR: (ezfb_clear_screen) FB NOT INITIALIZED\n");
        fflush(stderr);
    }
#endif
EZFB_FUNCTION_RETURN(fb->exists)
}

//############################################################################
int ezfb_swap_screens(struct ezfb* fb)
{
EZFB_FUNCTION_CALL
    u_char**   temp_row_uchar       = fb->screen_row_uchar     ;
    u_short**  temp_row_ushort      = fb->screen_row_ushort    ;
    u_int**    temp_row_uint        = fb->screen_row_uint      ;
    fb->screen_row_uchar            = fb->off_screen_row_uchar ;
    fb->screen_row_ushort           = fb->off_screen_row_ushort; 
    fb->screen_row_uint             = fb->off_screen_row_uint  ;   
    fb->off_screen_row_uchar        = temp_row_uchar           ;
    fb->off_screen_row_ushort       = temp_row_ushort          ;
    fb->off_screen_row_uint         = temp_row_uint            ;
EZFB_FUNCTION_RETURN(1)
}

//############################################################################
// ezfb destructor
//----------------------------------------------------------------------------
int ezfb_release(struct ezfb* fb)
{
EZFB_FUNCTION_CALL_0
    if(fb)
    {
        if(fb->exists)
        {
            ret = 1;
#if EZFB_API_MESSAGES
            printf("ezfb %s shutting down\n", fb->dev_name);
            fflush(stdout);
#endif      
            //----------------------------------------------------------------
            if(fb->bpp_was_changed)
            {
                ret &= ezfb_set_bpp(fb, fb->bpp_was);
#if EZFB_API_MESSAGES
                if(ret)
                {
                    printf("ezfb %s bpp set back to %d\n", fb->dev_name, fb->Var.bits_per_pixel);
                    fflush(stdout);
                }
#endif
            }
            //----------------------------------------------------------------
            if(fb->resolution_changed)
            {
                ret &= ezfb_set_xy_virtual(fb, fb->xres_virtual_was, fb->yres_virtual_was);
                ret &= ezfb_set_xy_offsets(fb, fb->xoffset_was     , fb->yoffset_was     );
#if EZFB_API_MESSAGES
                if(ret)
                {
                    printf("ezfb %s xres_virtual set back to %d\n", fb->dev_name, fb->Var.xres_virtual);
                    fflush(stdout);
                    printf("ezfb %s yres_virtual set back to %d\n", fb->dev_name, fb->Var.yres_virtual);
                    fflush(stdout);
                    printf("ezfb %s xoffset set back to %d\n"     , fb->dev_name, fb->Var.xoffset     );
                    fflush(stdout);
                    printf("ezfb %s yoffset set back to %d\n"     , fb->dev_name, fb->Var.yoffset     );
                    fflush(stdout);
                }
#endif
            }
            //----------------------------------------------------------------
            switch(fb->what_was_saved)
            {
                //------------------------------------------------------------
                case EZFB_SAVE_MEM:        ezfb_restore_mem(fb);
                                        break;
                //------------------------------------------------------------
                case EZFB_SAVE_SCREEN:    ezfb_restore_screen(fb);
                                        break;
                //------------------------------------------------------------
            }
            //----------------------------------------------------------------
            if(fb->mem_row_uchar      )    free(fb->mem_row_uchar      );
            if(fb->mem_row_uint       )    free(fb->mem_row_uint       );
            if(fb->mem_row_ushort     )    free(fb->mem_row_ushort     );
            if(fb->mem_row_uchar_xoff )    free(fb->mem_row_uchar_xoff );
            if(fb->mem_row_uint_xoff  )    free(fb->mem_row_uint_xoff  );
            if(fb->mem_row_ushort_xoff)    free(fb->mem_row_ushort_xoff);
#if EZFB_API_MESSAGES
            printf("ezfb %s display and screen matrix destroyed.\n", fb->dev_name);
            fflush(stdout);
#endif
            //----------------------------------------------------------------
            ret &= ezfb_release_cmap(fb);

#if !EZFB_NO_TTY0
            if(fb->dev_num == 0)
                ret &= ezfb_tty_text();
#endif

            //----------------------------------------------------------------
            close(fb->fd);
#if EZFB_API_MESSAGES
            printf("ezfb %s file descriptor closed.\n", fb->dev_name);
            fflush(stdout);
#endif
            fb->exists = 0;
            valid_ezfb[fb->dev_num] = NULL;
            num_valid_ezfb--;
#if EZFB_API_MESSAGES
            if(ret)
            {
                printf("ezfb %s SUCCESSFULLY RELEASED.\n", fb->dev_name);
                fflush(stdout);
            }
#endif
        }
        //--------------------------------------------------------------------
#if EZFB_API_ERRORS
        else
        {
            fprintf(stderr, "ezfb ERROR: (ezfb_release) FB NOT INITIALIZED.\n");
            fflush(stderr);
        }
#endif
    }
    //------------------------------------------------------------------------
#if EZFB_API_ERRORS
    else
    {
        fprintf(stderr, "ezfb ERROR: RELEASE CALLED ON NULL POINTER\n");
        fflush(stderr);
    }
#endif
EZFB_FUNCTION_RETURN(ret)
}

//############################################################################
int ezfb_dump_fb_fix_screeninfo(struct ezfb* fb)
{
EZFB_FUNCTION_CALL
    if(fb->exists)
    {
#if EZFB_API_MESSAGES
        printf("\n"                                                                   );
        printf("      FB FIXED SCREEN INFO:\n"                                        );
        printf("      fb->Fix.id               = %s\n"  ,        fb->Fix.id           ); // identification string eg "TT Builtin"
        printf("      fb->Fix.smem_start       = 0x%X\n", (u_int)fb->Fix.smem_start   ); // Start of frame buffer mem (physical address)
        printf("      fb->Fix.smem_len         = %u\n"  ,        fb->Fix.smem_len     ); // Length of frame buffer mem
        printf("      fb->Fix.type             = %u\n"  ,        fb->Fix.type         ); // see FB_TYPE_*        */
        printf("      fb->Fix.type_aux         = %u\n"  ,        fb->Fix.type_aux     ); // Interleave for interleaved Planes
        printf("      fb->Fix.visual           = %u\n"  ,        fb->Fix.visual       ); // see FB_VISUAL_*        */
        printf("      fb->Fix.xpanstep         = %u\n"  ,        fb->Fix.xpanstep     ); // zero if no hardware panning
        printf("      fb->Fix.ypanstep         = %u\n"  ,        fb->Fix.ypanstep     ); // zero if no hardware panning
        printf("      fb->Fix.ywrapstep        = %u\n"  ,        fb->Fix.ywrapstep    ); // zero if no hardware ywrap
        printf("      fb->Fix.line_length      = %u\n"  ,        fb->Fix.line_length  ); // length of a line in bytes
        printf("      fb->Fix.mmio_start       = 0x%X\n", (u_int)fb->Fix.mmio_start   ); // Start of Memory Mapped I/O (physical address)
        printf("      fb->Fix.mmio_len         = %u\n"  ,        fb->Fix.mmio_len     ); // Length of Memory Mapped I/O
        printf("      fb->Fix.accel            = %u\n"  ,        fb->Fix.accel        ); // Type of acceleration available
        printf("      fb->Fix.reserved[0]      = %u\n"  ,        fb->Fix.reserved[0]  ); // Reserved for future compatibility
        printf("      fb->Fix.reserved[1]      = %u\n"  ,        fb->Fix.reserved[1]  ); // Reserved for future compatibility
        printf("      fb->Fix.reserved[2]      = %u\n"  ,        fb->Fix.reserved[2]  ); // Reserved for future compatibility
        printf("\n");
        fflush(stdout);
#endif
    }
#if EZFB_API_ERRORS
    else
    {
        fprintf(stderr, "ezfb ERROR: (ezfb_dump_fb_fix_screeninfo) FB NOT INITIALIZED\n");
        fflush(stderr);
    }
#endif
EZFB_FUNCTION_RETURN(fb->exists)
}

//############################################################################
int ezfb_dump_fb_var_screeninfo(struct ezfb* fb)
{
EZFB_FUNCTION_CALL
#if EZFB_API_MESSAGES
    if(fb->exists)
    {
        ezfb_get_fb_var(fb);
        printf("\n"                                                                );
        printf("      FB VARIABLE SCREEN INFO:\n"                                  );
        printf("      fb->Var.xres             = %u\n"  , fb->Var.xres             ); // visible resolution
        printf("      fb->Var.yres             = %u\n"  , fb->Var.yres             );
        printf("      fb->Var.xres_virtual     = %u\n"  , fb->Var.xres_virtual     ); // virtual resolution
        printf("      fb->Var.yres_virtual     = %u\n"  , fb->Var.yres_virtual     );
        printf("      fb->Var.xoffset          = %u\n"  , fb->Var.xoffset          ); // offset from virtual to visible
        printf("      fb->Var.yoffset          = %u\n"  , fb->Var.yoffset          ); // resolution
        printf("      fb->Var.bits_per_pixel   = %u\n"  , fb->Var.bits_per_pixel   ); // guess what
        printf("      fb->Var.grayscale        = %u\n"  , fb->Var.grayscale        ); // != 0 Graylevels instead of colors
        printf("      fb->Var.red.offset       = %u\n"  , fb->Var.red.offset       ); // bitfield in fb mem if true color,
        printf("      fb->Var.red.length       = %u\n"  , fb->Var.red.length       ); // else only length is significant
        printf("      fb->Var.red.msb_right    = %u\n"  , fb->Var.red.msb_right    );
        printf("      fb->Var.green.offset     = %u\n"  , fb->Var.green.offset     );
        printf("      fb->Var.green.length     = %u\n"  , fb->Var.green.length     );
        printf("      fb->Var.green.msb_right  = %u\n"  , fb->Var.green.msb_right  );
        printf("      fb->Var.blue.offset      = %u\n"  , fb->Var.blue.offset      );
        printf("      fb->Var.blue.length      = %u\n"  , fb->Var.blue.length      );
        printf("      fb->Var.blue.msb_right   = %u\n"  , fb->Var.blue.msb_right   );
        printf("      fb->Var.transp.offset    = %u\n"  , fb->Var.transp.offset    ); // transparency
        printf("      fb->Var.transp.length    = %u\n"  , fb->Var.transp.length    );
        printf("      fb->Var.transp.msb_right = %u\n"  , fb->Var.transp.msb_right );
        printf("      fb->Var.nonstd           = %u\n"  , fb->Var.nonstd           ); // != 0 Non standard pixel format
        printf("      fb->Var.activate         = %u\n"  , fb->Var.activate         ); // see FB_ACTIVATE_*
        printf("      fb->Var.height           = %u\n"  , fb->Var.height           ); // height of picture in mm
        printf("      fb->Var.width            = %u\n"  , fb->Var.width            ); // width of picture in mm
        printf("      fb->Var.accel_flags      = %u\n"  , fb->Var.accel_flags      ); // acceleration flags (hints)
        printf("      fb->Var.pixclock         = %u\n"  , fb->Var.pixclock         ); // pixel clock in ps (pico seconds)
        printf("      fb->Var.left_margin      = %u\n"  , fb->Var.left_margin      ); // time from sync to picture
        printf("      fb->Var.right_margin     = %u\n"  , fb->Var.right_margin     ); // time from picture to sync
        printf("      fb->Var.upper_margin     = %u\n"  , fb->Var.upper_margin     ); // time from sync to picture
        printf("      fb->Var.lower_margin     = %u\n"  , fb->Var.lower_margin     );
        printf("      fb->Var.hsync_len        = %u\n"  , fb->Var.hsync_len        ); // length of horizontal sync
        printf("      fb->Var.vsync_len        = %u\n"  , fb->Var.vsync_len        ); // length of vertical sync
        printf("      fb->Var.sync             = %u\n"  , fb->Var.sync             ); // see FB_SYNC_*
        printf("      fb->Var.vmode            = %u\n"  , fb->Var.vmode            ); // see FB_VMODE_*
        printf("      fb->Var.reserved[0]      = %u\n"  , fb->Var.reserved[0]      ); // Reserved for future compatibility
        printf("      fb->Var.reserved[1]      = %u\n"  , fb->Var.reserved[1]      ); // Reserved for future compatibility
        printf("      fb->Var.reserved[2]      = %u\n"  , fb->Var.reserved[2]      ); // Reserved for future compatibility
        printf("      fb->Var.reserved[3]      = %u\n"  , fb->Var.reserved[3]      ); // Reserved for future compatibility
        printf("      fb->Var.reserved[4]      = %u\n"  , fb->Var.reserved[4]      ); // Reserved for future compatibility
        printf("      fb->Var.reserved[5]      = %u\n"  , fb->Var.reserved[5]      ); // Reserved for future compatibility
        printf("\n");
        fflush(stdout);
    }
    else
    {
        fprintf(stderr, "ezfb ERROR: (ezfb_dump_fb_var_screeninfo) FB NOT INITIALIZED\n");
        fflush(stderr);
    }
#endif
EZFB_FUNCTION_RETURN(fb->exists)
}

//############################################################################
int ezfb_dump(struct ezfb* fb)
{
EZFB_FUNCTION_CALL
#if EZFB_API_MESSAGES
    if(fb->exists)
    {
        printf("      EZFB GLOBAL PARAMETERS:\n"                                  );
        printf("      fb->fd                 = %d\n"  , fb->fd                    );
        printf("      fb->dev_num            = %u\n"  , fb->dev_num               );
        printf("      fb->dev_name           = %s\n"  , fb->dev_name              );
        printf("      fb->exists             = %u\n"  , fb->exists                );
        printf("      fb->resolution_changed = %u\n"  , fb->resolution_changed    );
        printf("      fb->bpp_was_changed    = %u\n"  , fb->bpp_was_changed       );
        printf("      fb->xres_was           = %u\n"  , fb->xres_was              );
        printf("      fb->yres_was           = %u\n"  , fb->yres_was              );
        printf("      fb->xres_virtual_was   = %u\n"  , fb->xres_virtual_was      );
        printf("      fb->yres_virtual_was   = %u\n"  , fb->yres_virtual_was      );
        printf("      fb->xoffset_was        = %u\n"  , fb->xoffset_was           );
        printf("      fb->yoffset_was        = %u\n"  , fb->yoffset_was           );
        printf("      fb->bpp_was            = %u\n"  , fb->bpp_was               );
        printf("      fb->mem_uchar          = 0x%X\n", (u_int)fb->mem_uchar      );
        printf("      fb->data_was           = 0x%X\n", (u_int)fb->data_was       );
        printf("      fb->max_x              = %u\n"  , fb->max_x                 );
        printf("      fb->max_y              = %u\n"  , fb->max_y                 );
        printf("      fb->mid_x              = %u\n"  , fb->mid_x                 );
        printf("      fb->mid_y              = %u\n"  , fb->mid_y                 );
        printf("\n");
        fflush(stdout);
    }
    else
    {
        fprintf(stderr, "ezfb ERROR: (ezfb_dump) FB NOT INITIALIZED\n");
        fflush(stderr);
    }
#endif
EZFB_FUNCTION_RETURN(fb->exists)
}

//############################################################################
int ezfb_dump_screen_map(struct ezfb* fb)
{
    u_int x, y;
EZFB_FUNCTION_CALL
#if EZFB_API_MESSAGES
    if(fb->exists)
    {
        printf("\nFB SCREEN MAP %d x %d : %d pixels:", fb->Var.xres, fb->Var.yres, fb->Var.xres * fb->Var.yres);
        switch(fb->Var.bits_per_pixel)
        {
            case  1:    for(y = 0; y < fb->Var.yres; y++)
                        {
                            for(x = 0; x < (fb->Var.xres / 8); x++)
                                printf(" %2X" , fb->screen_row_uchar[y][x]);
                            printf("\n");
                            fflush(stdout);
                        }
            //--------------------------------------------------------------------
            case  2:    for(y = 0; y < fb->Var.yres; y++)
                        {
                            for(x = 0; x < (fb->Var.xres / 4); x++)
                                printf(" %2X" , fb->screen_row_uchar[y][x]);
                            printf("\n");
                            fflush(stdout);
                        }
            //--------------------------------------------------------------------
            case  4:    for(y = 0; y < fb->Var.yres; y++)
                        {
                            for(x = 0; x < (fb->Var.xres / 2); x++)
                                printf(" %2X" , fb->screen_row_uchar[y][x]);
                            printf("\n");
                            fflush(stdout);
                        }
            //--------------------------------------------------------------------
            case  8:    for(y = 0; y < fb->Var.yres; y++)
                        {
                            for(x = 0; x < fb->Var.xres; x++)
                                printf(" %2X" , fb->screen_row_uchar[y][x]);
                            printf("\n");
                            fflush(stdout);
                        }
            //--------------------------------------------------------------------
            case 15:
            case 16: for(y = 0; y < fb->Var.yres; y++)
                        {
                            for(x = 0; x < fb->Var.xres; x++)
                                printf(" %4X" , fb->screen_row_ushort[y][x]);
                            printf("\n");
                            fflush(stdout);
                        }
            //--------------------------------------------------------------------
            case 24: for(y = 0; y < fb->Var.yres; y++)
                        {
                            for(x = 0; x < (fb->Var.xres * 3); x+=3)
                            {
                                printf(" [%2X" , fb->screen_row_uchar[y][x    ]);
                                printf(",%2X"  , fb->screen_row_uchar[y][x + 1]);
                                printf(",%2X]" , fb->screen_row_uchar[y][x + 2]);
                            }
                            printf("\n");
                            fflush(stdout);
                        }
            //--------------------------------------------------------------------
            case 32: for(y = 0; y < fb->Var.yres; y++)
                        {
                            for(x = 0; x < fb->Var.xres; x++)
                                printf(" %8X" , fb->screen_row_uint[y][x]);
                            printf("\n");
                            fflush(stdout);
                        }
            //--------------------------------------------------------------------
        }
    }
    else
    {
        fprintf(stderr, "ezfb ERROR: (ezfb_dump_screen_map) FB NOT INITIALIZED\n");
        fflush(stderr);
    }
#endif
EZFB_FUNCTION_RETURN(fb->exists)
}

//############################################################################
int ezfb_dump_virtual_map(struct ezfb* fb)
{
    u_int x, y;
EZFB_FUNCTION_CALL
#if EZFB_API_MESSAGES
    if(fb->exists)
    {
        printf("\nFB VIRTUAL MAP %d x %d : %d pixels:", fb->Var.xres_virtual, fb->Var.yres_virtual, fb->Var.xres_virtual * fb->Var.yres_virtual);
        switch(fb->Var.bits_per_pixel)
        {
            case  1:    for(y = 0; y < fb->Var.yres_virtual; y++)
                        {
                            for(x = 0; x < (fb->Var.xres_virtual / 8); x++)
                                printf(" %2X" , fb->mem_row_uchar[y][x]);
                            printf("\n");
                            fflush(stdout);
                        }
            //--------------------------------------------------------------------
            case  2:    for(y = 0; y < fb->Var.yres_virtual; y++)
                        {
                            for(x = 0; x < (fb->Var.xres_virtual / 4); x++)
                                printf(" %2X" , fb->mem_row_uchar[y][x]);
                            printf("\n");
                            fflush(stdout);
                        }
            //--------------------------------------------------------------------
            case  4:    for(y = 0; y < fb->Var.yres_virtual; y++)
                        {
                            for(x = 0; x < (fb->Var.xres_virtual / 2); x++)
                                printf(" %2X" , fb->mem_row_uchar[y][x]);
                            printf("\n");
                            fflush(stdout);
                        }
            //--------------------------------------------------------------------
            case  8:    for(y = 0; y < fb->Var.yres_virtual; y++)
                        {
                            for(x = 0; x < fb->Var.xres_virtual; x++)
                                printf(" %2X" , fb->mem_row_uchar[y][x]);
                            printf("\n");
                            fflush(stdout);
                        }
            //--------------------------------------------------------------------
            case 15:
            case 16:    for(y = 0; y < fb->Var.yres_virtual; y++)
                        {
                            for(x = 0; x < fb->Var.xres_virtual; x++)
                                printf(" %4X" , fb->mem_row_ushort[y][x]);
                            printf("\n");
                            fflush(stdout);
                        }
            //--------------------------------------------------------------------
            case 24:    for(y = 0; y < fb->Var.yres_virtual; y++)
                        {
                            for(x = 0; x < (fb->Var.xres_virtual * 3); x+=3)
                            {
                                printf(" [%2X" , fb->mem_row_uchar[y][x    ]);
                                printf(",%2X"  , fb->mem_row_uchar[y][x + 1]);
                                printf(",%2X]" , fb->mem_row_uchar[y][x + 2]);
                            }
                            printf("\n");
                            fflush(stdout);
                        }
            //--------------------------------------------------------------------
            case 32:    for(y = 0; y < fb->Var.yres_virtual; y++)
                        {
                            for(x = 0; x < fb->Var.xres_virtual; x++)
                                printf(" %8X" , fb->mem_row_uint[y][x]);
                            printf("\n");
                            fflush(stdout);
                        }
            //--------------------------------------------------------------------
        }
    }
    else
    {
        fprintf(stderr, "ezfb ERROR: (ezfb_dump_virtual_map) FB NOT INITIALIZED\n");
        fflush(stderr);
    }
#endif
EZFB_FUNCTION_RETURN(fb->exists)
}

//############################################################################
int ezfb_binfile_screen_map(struct ezfb* fb, char* file_name)
{
EZFB_FUNCTION_CALL_0
    if(fb->exists)
    {
        int binfile_fd;
        if(0 > (binfile_fd = open(file_name, (O_WRONLY | O_CREAT), (S_IRWXU | S_IRGRP | S_IROTH)))) // mode 744
        {
#if EZFB_API_ERRORS
            perror("ezfb ERROR: opening output file failed\n");
#endif
            ret = 0;
        }
        else
        {
            ret = 1;
            write(binfile_fd, (void*)(&(fb->screen_row_uchar[0][0])), fb->Var.xres * fb->Var.yres);
            close(binfile_fd);
        }
    }
#if EZFB_API_ERRORS
    else
    {
        fprintf(stderr, "ezfb ERROR: (ezfb_binfile_screen_map) FB NOT INITIALIZED\n");
        fflush(stderr);
    }
#endif
EZFB_FUNCTION_RETURN(ret)
}

//############################################################################
int ezfb_binfile_mem_map(struct ezfb* fb, char* file_name)
{
EZFB_FUNCTION_CALL_0
    if(fb->exists)
    {
        int binfile_fd;
        if(0 > (binfile_fd = open(file_name, (O_WRONLY | O_CREAT), (S_IRWXU | S_IRGRP | S_IROTH)))) // mode 744
        {
#if EZFB_API_ERRORS
            perror("ezfb ERROR: opening output file failed\n");
#endif
            ret = 0;
        }
        else
        {
            ret = 1;
            write(binfile_fd, (void*)fb->mem_uchar, fb->Fix.smem_len);
            close(binfile_fd);
        }
    }
#if EZFB_API_ERRORS
    else
    {
        fprintf(stderr, "ezfb ERROR: (ezfb_binfile_mem_map) FB NOT INITIALIZED\n");
        fflush(stderr);
    }
#endif
EZFB_FUNCTION_RETURN(ret)
}

//############################################################################
float ezfb_random_01() // returns a float between 0 & 1.
{
    return ((float)(rand()) / (float)RAND_MAX);
}

//############################################################################
float ezfb_random_neg_to_pos_1() // returns a float between -1 & 1.
{
    return ezfb_random_01() - ezfb_random_01();
}

//############################################################################
void ezfb_set_functions(struct ezfb* fb)
{
EZFB_FUNCTION_CALL
    switch(fb->Var.bits_per_pixel)
    {
        // -------------------------------------------------------------
        case  1: fb->make_rgb  = ezfb_make_rgb__1bit  ;
                 fb->put_pixel = ezfb_put_pixel__1bit ;
                 fb->get_pixel = ezfb_get_pixel__1bit ;
                 break;
        // -------------------------------------------------------------
        case  2: fb->make_rgb  = ezfb_make_rgb__1bit  ;
                 fb->put_pixel = ezfb_put_pixel__2bit ;
                 fb->get_pixel = ezfb_get_pixel__2bit ;
                 break;
        // -------------------------------------------------------------
        case  4: fb->make_rgb  = ezfb_make_rgb__4bit  ;
                 fb->put_pixel = ezfb_put_pixel__4bit ;
                 fb->get_pixel = ezfb_get_pixel__4bit ;
                 break;
        // -------------------------------------------------------------
        case  8: fb->make_rgb  = ezfb_make_rgb__8bit  ;
                 fb->put_pixel = ezfb_put_pixel__8bit ;
                 fb->get_pixel = ezfb_get_pixel__8bit ;
                 break;
        // -------------------------------------------------------------
        case 15:
        case 16: fb->make_rgb  = ezfb_make_rgb_16bit  ;
                 fb->put_pixel = ezfb_put_pixel_16bit ;
                 fb->get_pixel = ezfb_get_pixel_16bit ;
                 break;
        // -------------------------------------------------------------
        case 24: fb->make_rgb  = ezfb_make_rgb_32bit  ;
                 fb->put_pixel = ezfb_put_pixel_24bit ;
                 fb->get_pixel = ezfb_get_pixel_24bit ;
                 break;
        // -------------------------------------------------------------
        case 32: fb->make_rgb  = ezfb_make_rgb_32bit  ;
                 fb->put_pixel = ezfb_put_pixel_32bit ;
                 fb->get_pixel = ezfb_get_pixel_32bit ;
                 break;
        // -------------------------------------------------------------
        default: exit(55);
        // -------------------------------------------------------------
    }
EZFB_FUNCTION_RETURN_VOID
}

//############################################################################
void ezfb_set_sig_traps()
{
EZFB_FUNCTION_CALL
    static int been_done = 0;
    if(!been_done)
    {
        signal(SIGTERM, ezfb_terminate);
        signal(SIGINT , ezfb_terminate);
        signal(SIGQUIT, ezfb_terminate);
        signal(SIGHUP , ezfb_terminate);
        signal(SIGTSTP, ezfb_terminate);
        signal(SIGTTIN, ezfb_terminate);
        signal(SIGFPE , ezfb_terminate);
        signal(SIGILL , ezfb_terminate);
        signal(SIGSEGV, ezfb_terminate);
        signal(SIGBUS , ezfb_terminate);
        signal(SIGABRT, ezfb_terminate);
        signal(SIGSYS , ezfb_terminate);
        been_done = 1;
#if EZFB_API_MESSAGES
        printf("TERMINATION SIGNAL TRAPS SET\n");
        fflush(stdout);
#endif
    }
EZFB_FUNCTION_RETURN_VOID
}

//############################################################################
void ezfb_terminate(int signum)
{
    int i;
    for(i = 0; i <= EZFB_MAX_FB_NUM; i++)
    {
        if(valid_ezfb[i])
        {
#if EZFB_API_MESSAGES
            printf(   "FRAME BUFFER %s TERMINATING WITH SIGNUM : %d\n\n"
                    , valid_ezfb[i]->dev_name
                    , signum
                  );
            fflush(stdout);
#endif
            ezfb_release(valid_ezfb[i]);
#if EZFB_API_MESSAGES
            printf("DONE\n");
            fflush(stdout);
#endif
        }
    }
    printf("\n");
    fflush(stdout);    
    exit(signum);
}

//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################

