//############################################################################
//
// display.cpp
//
// ~  This file is part of the LaserBoy package   ~
// ~   International Laser Display Association    ~
// ~ ILDA file display and transformations in C++ ~
// ~               built for EZFB                 ~
//
// by James Lehman
// Extra Stimulus Inc.
// james@akrobiz.com
//
// began : October 2003
//
//############################################################################
#include "ez_ilda.hpp"
#include "ez_wave.hpp"
#include "ez_oss_dsp.hpp"

//############################################################################
ifstream                 in;

string                   file("laserboy.wav");

struct  ezfb             fb             = {0};
struct  ezfb_bitmap      bmp            = {0};

int                      i                            ,
                         error                        ,
                         span                         ,
                         last                         ,
                         red                          ,
                         green                        ,
                         blue                         ,
                         y_flip_flop    =  0          ,
                         sample_count   =  0          ,
                         frame_count    =  0          ,
                         offset_max     = -0x80000000 ,
                         offset_min     =  0x7FFFFFFF ;

bool                     show_points    = false ,
                         show_blanks    = false ,
                         real_time      = false ,
                         inverted       = false ,
                         no_blank       = false ,
                         no_ezfb_buffer = false ;

unsigned long long int   time_zero        ,
                         ticks_per_sample ;

float                    scale ;

int                      color = 0x00ffffff ,
                         fb_center          ;

ez_wave_header           header ;

typedef ez_wave_sample*  ez_wave_sample_pointer ;
ez_wave_sample_pointer   roll_over              ;
ez_wave_sample_pointer*  sample_window          ;

//############################################################################
void read_args(int argc, char* argv[])
{
    if(argc > 1) // file name
    {
        file = argv[1];
        if(    file.size() < 4
            || file.substr(file.size() - 4) != ".wav"
          )
            file += ".wav";
    }
    if(argc > 2)  show_points    = (bool)atoi(argv[2]);
    if(argc > 3)  show_blanks    = (bool)atoi(argv[3]);
    if(argc > 4)  real_time      = (bool)atoi(argv[4]);
    if(argc > 5)  inverted       = (bool)atoi(argv[5]);
    if(argc > 6)  no_blank       = (bool)atoi(argv[6]);
    if(argc > 7)  no_ezfb_buffer = (bool)atoi(argv[7]);

    cout << endl;
    cout << "show_points     : " << ((show_points   ) ? ("yes") : ("no")) << endl;
    cout << "show_blanks     : " << ((show_blanks   ) ? ("yes") : ("no")) << endl;
    cout << "real_time       : " << ((real_time     ) ? ("yes") : ("no")) << endl;
    cout << "inverted        : " << ((inverted      ) ? ("yes") : ("no")) << endl;
    cout << "no_blank        : " << ((no_blank      ) ? ("yes") : ("no")) << endl;
    cout << "no_ezfb_buffer  : " << ((no_ezfb_buffer) ? ("yes") : ("no")) << endl;
    return;
}

//############################################################################
bool setup_ezfb()
{
    bool can_do_double_buffer = false;
    if(!ezfb_init(&fb, EZFB_SAVE_SCREEN))
        exit(-1);
    //------------------------------------------------------------------------
	if(fb.Var.bits_per_pixel <= 8)
		if(!ezfb_set_bpp(&fb, 16))
		{
			cout << "16 bit frame buffer not supported" << endl;
			if(!ezfb_set_bpp(&fb, 32))
			{
				cout << "32 bit frame buffer not supported" << endl;
				if(!ezfb_set_bpp(&fb, 24))
				{
					cout << "24 bit frame buffer not supported" << endl;
					exit(-1);
				}
			}
		}
    //------------------------------------------------------------------------
    scale     = fb.Var.yres / 65536.0;
    fb_center = fb.Var.yres / 2;
    //------------------------------------------------------------------------
	if(no_ezfb_buffer)
		ezfb_set_xy_virtual(&fb, fb.Var.xres, fb.Var.yres);
	else
	{
		ezfb_set_xy_virtual(&fb, fb.Var.xres, (fb.Var.yres * 2));
		can_do_double_buffer = (fb.Var.yres_virtual >= (fb.Var.yres * 2));
		if(!can_do_double_buffer)
			ezfb_set_xy_virtual(&fb, fb.Var.xres, fb.Var.yres);
	}
    //------------------------------------------------------------------------
    ezfb_clear_mem(&fb);
    //------------------------------------------------------------------------
    return can_do_double_buffer;
}

//############################################################################
void open_wave()
{
    in.open(file.c_str(), ios::in | ios::binary);
    if(!in.is_open())
    {
        cout << endl << file << " failed to open for input" << endl;
        exit(-1);
    }
    //------------------------------------------------------------------------
    if((error = header.from_ifstream(in)))
    {
        cout << endl << "bad wave file : " << error << endl << endl;
        exit(-1);
    }
    header.tell();
    //------------------------------------------------------------------------
    ticks_per_sample = (unsigned long long int)(cpu_speed() / header.sample_rate);
    //------------------------------------------------------------------------
    for(i = 0; i < header.num_channels; i++)
    {
        if(header.offset[i] > offset_max)
            offset_max = header.offset[i];
        if(header.offset[i] < offset_min)
            offset_min = header.offset[i];
    }
    //------------------------------------------------------------------------
    span  = (offset_max - offset_min) + 2,
    last  = span - 1,
    red   = -header.offset[2] + 1,
    green = -header.offset[3] + 1,
    blue  = -header.offset[4] + 1;
    //------------------------------------------------------------------------
#if EZ_ILDA_DEBUG_MESSAGES
    cout << "span  : " << span   << endl
         << "red   : " << red    << endl
         << "green : " << green  << endl
         << "blue  : " << blue   << endl  << endl;
#endif
    //------------------------------------------------------------------------
    sample_window = new ez_wave_sample_pointer[span];
    for(i = 0; i < span; i++)
        sample_window[i] = new ez_wave_sample(header.num_channels);
    return;
}

//############################################################################
void ezfb_buffer() // noninverted double buffer in video ram
{
    time_zero = rdtsc();
    //-----------------------------------------------------------------------
    while(true)
    {
        roll_over = sample_window[0];
        for(i = 0; i < last; i++)
            sample_window[i] = sample_window[i + 1];
        sample_window[last] = roll_over;

        if(sample_window[last]->from_ifstream(in))
            return;
        //-------------------------------------------------------------------
        sample_count++;
        color = fb.make_rgb(   &fb
                             , sample_window[red  ]->channel[2] / 128
                             , sample_window[green]->channel[3] / 128
                             , sample_window[blue ]->channel[4] / 128
                           );
        //-------------------------------------------------------------------
        if(color)
        {
            ezfb_put_line (   &fb
                            , (u_int)((sample_window[0]->channel[0] *  scale) + fb_center              )
                            , (u_int)((sample_window[0]->channel[1] * -scale) + fb_center + y_flip_flop)
                            , (u_int)((sample_window[1]->channel[0] *  scale) + fb_center              )
                            , (u_int)((sample_window[1]->channel[1] * -scale) + fb_center + y_flip_flop)
                            , color
                          );

            if(show_points)
                ezfb_put_rectangle(   &fb
                                    , (u_int)((sample_window[1]->channel[0] *  scale) + fb_center - 1              )
                                    , (u_int)((sample_window[1]->channel[1] * -scale) + fb_center - 1 + y_flip_flop)
                                    , (u_int)((sample_window[1]->channel[0] *  scale) + fb_center + 1              )
                                    , (u_int)((sample_window[1]->channel[1] * -scale) + fb_center + 1 + y_flip_flop)
                                    , color
                                    , false
                                  );
        }
        //-------------------------------------------------------------------
        else if(show_blanks)
        {
            ezfb_put_line (   &fb
                            , (u_int)((sample_window[0]->channel[0] *  scale) + fb_center              )
                            , (u_int)((sample_window[0]->channel[1] * -scale) + fb_center + y_flip_flop)
                            , (u_int)((sample_window[1]->channel[0] *  scale) + fb_center              )
                            , (u_int)((sample_window[1]->channel[1] * -scale) + fb_center + y_flip_flop)
                            , fb.colors.gray
                          );

            if(show_points)
                ezfb_put_rectangle(   &fb
                                    , (u_int)((sample_window[1]->channel[0] *  scale) + fb_center - 1              )
                                    , (u_int)((sample_window[1]->channel[1] * -scale) + fb_center - 1 + y_flip_flop)
                                    , (u_int)((sample_window[1]->channel[0] *  scale) + fb_center + 1              )
                                    , (u_int)((sample_window[1]->channel[1] * -scale) + fb_center + 1 + y_flip_flop)
                                    , fb.colors.gray
                                    , false
                                  );
        }
        //-------------------------------------------------------------------
        if(sample_window[red]->channel[2] & 0x0001)
        {
            if(real_time)
                do{;} while(   sample_count * ticks_per_sample
                             > rdtsc() - time_zero
                           );

            ezfb_set_xy_offsets(&fb, 0, y_flip_flop);

            if(y_flip_flop)
                y_flip_flop = 0;
            else
                y_flip_flop = fb.Var.yres;

            if(!no_blank)
                ezfb_clear_screen_at_offsets(&fb, 0, y_flip_flop);

            frame_count++;
        }
        //-------------------------------------------------------------------
    }
    return;
}

//############################################################################
void ezfb_buffer_inverted() // inverted double buffer in video ram
{
    time_zero = rdtsc();
    //-----------------------------------------------------------------------
    while(true)
    {
        roll_over = sample_window[0];
        for(i = 0; i < last; i++)
            sample_window[i] = sample_window[i + 1];
        sample_window[last] = roll_over;

        if(sample_window[last]->from_ifstream(in))
            return;
        //-------------------------------------------------------------------
        sample_count++;
        color = fb.make_rgb(   &fb
                             , 256 - (sample_window[red  ]->channel[2] / 128)
                             , 256 - (sample_window[green]->channel[3] / 128)
                             , 256 - (sample_window[blue ]->channel[4] / 128)
                           );
        //-------------------------------------------------------------------
        if(color)
        {
            ezfb_put_line (   &fb
                            , (u_int)((sample_window[0]->channel[0] * -scale) + fb_center              )
                            , (u_int)((sample_window[0]->channel[1] *  scale) + fb_center + y_flip_flop)
                            , (u_int)((sample_window[1]->channel[0] * -scale) + fb_center              )
                            , (u_int)((sample_window[1]->channel[1] *  scale) + fb_center + y_flip_flop)
                            , color
                          );

            if(show_points)
                ezfb_put_rectangle(   &fb
                                    , (u_int)((sample_window[1]->channel[0] * -scale) + fb_center - 1              )
                                    , (u_int)((sample_window[1]->channel[1] *  scale) + fb_center - 1 + y_flip_flop)
                                    , (u_int)((sample_window[1]->channel[0] * -scale) + fb_center + 1              )
                                    , (u_int)((sample_window[1]->channel[1] *  scale) + fb_center + 1 + y_flip_flop)
                                    , color
                                    , false
                                  );
        }
        //-------------------------------------------------------------------
        else if(show_blanks)
        {
            ezfb_put_line (   &fb
                            , (u_int)((sample_window[0]->channel[0] * -scale) + fb_center              )
                            , (u_int)((sample_window[0]->channel[1] *  scale) + fb_center + y_flip_flop)
                            , (u_int)((sample_window[1]->channel[0] * -scale) + fb_center              )
                            , (u_int)((sample_window[1]->channel[1] *  scale) + fb_center + y_flip_flop)
                            , fb.colors.gray
                          );

            if(show_points)
                ezfb_put_rectangle(   &fb
                                    , (u_int)((sample_window[1]->channel[0] * -scale) + fb_center - 1              )
                                    , (u_int)((sample_window[1]->channel[1] *  scale) + fb_center - 1 + y_flip_flop)
                                    , (u_int)((sample_window[1]->channel[0] * -scale) + fb_center + 1              )
                                    , (u_int)((sample_window[1]->channel[1] *  scale) + fb_center + 1 + y_flip_flop)
                                    , fb.colors.gray
                                    , false
                                  );
        }
        //-------------------------------------------------------------------
        if(sample_window[red]->channel[2] & 0x0001)
        {
            if(real_time)
                do{;} while(   sample_count * ticks_per_sample
                             > rdtsc() - time_zero
                           );

            ezfb_set_xy_offsets(&fb, 0, y_flip_flop);

            if(y_flip_flop)
                y_flip_flop = 0;
            else
                y_flip_flop = fb.Var.yres;

            if(!no_blank)
                ezfb_clear_screen_at_offsets(&fb, 0, y_flip_flop);

            frame_count++;
        }
        //-------------------------------------------------------------------
    }
    return;
}

//############################################################################
void bmp_buffer() // noninverted buffer in system ram
{
    bmp_clone_ezfb(&fb, &bmp, 0);
    time_zero = rdtsc();
    //-----------------------------------------------------------------------
    while(true)
    {
        roll_over = sample_window[0];
        for(i = 0; i < last; i++)
            sample_window[i] = sample_window[i + 1];
        sample_window[last] = roll_over;

        if(sample_window[last]->from_ifstream(in))
            return;
        //-------------------------------------------------------------------
        sample_count++;
        color = fb.make_rgb(   &fb
                             , sample_window[red  ]->channel[2] / 128
                             , sample_window[green]->channel[3] / 128
                             , sample_window[blue ]->channel[4] / 128
                           );
        //-------------------------------------------------------------------
        if(color)
        {
            bmp_put_line_fast(   &bmp
                               , (u_int)((sample_window[0]->channel[0] *  scale) + fb_center)
                               , (u_int)((sample_window[0]->channel[1] * -scale) + fb_center)
                               , (u_int)((sample_window[1]->channel[0] *  scale) + fb_center)
                               , (u_int)((sample_window[1]->channel[1] * -scale) + fb_center)
                               , color
                             );

            if(show_points)
                bmp_put_rectangle(   &bmp
                                   , (u_int)((sample_window[1]->channel[0] *  scale) + fb_center - 1)
                                   , (u_int)((sample_window[1]->channel[1] * -scale) + fb_center - 1)
                                   , (u_int)((sample_window[1]->channel[0] *  scale) + fb_center + 1)
                                   , (u_int)((sample_window[1]->channel[1] * -scale) + fb_center + 1)
                                   , color
                                   , false
                                 );
        }
        //-------------------------------------------------------------------
        else if(show_blanks)
        {
            bmp_put_line_fast(   &bmp
                               , (u_int)((sample_window[0]->channel[0] *  scale) + fb_center)
                               , (u_int)((sample_window[0]->channel[1] * -scale) + fb_center)
                               , (u_int)((sample_window[1]->channel[0] *  scale) + fb_center)
                               , (u_int)((sample_window[1]->channel[1] * -scale) + fb_center)
                               , bmp.gray
                             );

            if(show_points)
                bmp_put_rectangle(   &bmp
                                   , (u_int)((sample_window[1]->channel[0] *  scale) + fb_center - 1)
                                   , (u_int)((sample_window[1]->channel[1] * -scale) + fb_center - 1)
                                   , (u_int)((sample_window[1]->channel[0] *  scale) + fb_center + 1)
                                   , (u_int)((sample_window[1]->channel[1] * -scale) + fb_center + 1)
                                   , bmp.gray
                                   , false
                                  );
        }
        //-------------------------------------------------------------------
        if(sample_window[red]->channel[2] & 0x0001)
        {
            if(real_time)
                do{;} while(   sample_count * ticks_per_sample
                             > rdtsc() - time_zero
                           );
            bmp_clone_to_ezfb(&fb, &bmp);
            if(!no_blank)
                bmp_clear(&bmp, 0x00);
            frame_count++;
        }
        //-------------------------------------------------------------------
    }
    return;
}

//############################################################################
void bmp_buffer_inverted() // inverted buffer in system ram
{
    bmp_clone_ezfb(&fb, &bmp, 0);
    time_zero = rdtsc();
    //-----------------------------------------------------------------------
    while(true)
    {
        roll_over = sample_window[0];
        for(i = 0; i < last; i++)
            sample_window[i] = sample_window[i + 1];
        sample_window[last] = roll_over;

        if(sample_window[last]->from_ifstream(in))
            return;
        //-------------------------------------------------------------------
        sample_count++;
        color = fb.make_rgb(   &fb
                             , 256 - (sample_window[red  ]->channel[2] / 128)
                             , 256 - (sample_window[green]->channel[3] / 128)
                             , 256 - (sample_window[blue ]->channel[4] / 128)
                           );
        //-------------------------------------------------------------------
        if(color)
        {
            bmp_put_line_fast(   &bmp
                               , (u_int)((sample_window[0]->channel[0] * scale) + fb_center)
                               , (u_int)((sample_window[0]->channel[1] * scale) + fb_center)
                               , (u_int)((sample_window[1]->channel[0] * scale) + fb_center)
                               , (u_int)((sample_window[1]->channel[1] * scale) + fb_center)
                               , color
                             );

            if(show_points)
                bmp_put_rectangle(   &bmp
                                   , (u_int)((sample_window[1]->channel[0] * scale) + fb_center - 1)
                                   , (u_int)((sample_window[1]->channel[1] * scale) + fb_center - 1)
                                   , (u_int)((sample_window[1]->channel[0] * scale) + fb_center + 1)
                                   , (u_int)((sample_window[1]->channel[1] * scale) + fb_center + 1)
                                   , color
                                   , false
                                 );
        }
        //-------------------------------------------------------------------
        else if(show_blanks)
        {
            bmp_put_line_fast(   &bmp
                               , (u_int)((sample_window[0]->channel[0] * scale) + fb_center)
                               , (u_int)((sample_window[0]->channel[1] * scale) + fb_center)
                               , (u_int)((sample_window[1]->channel[0] * scale) + fb_center)
                               , (u_int)((sample_window[1]->channel[1] * scale) + fb_center)
                               , bmp.gray
                             );

            if(show_points)
                bmp_put_rectangle(   &bmp
                                   , (u_int)((sample_window[1]->channel[0] * scale) + fb_center - 1)
                                   , (u_int)((sample_window[1]->channel[1] * scale) + fb_center - 1)
                                   , (u_int)((sample_window[1]->channel[0] * scale) + fb_center + 1)
                                   , (u_int)((sample_window[1]->channel[1] * scale) + fb_center + 1)
                                   , bmp.gray
                                   , false
                                 );
        }
        //-------------------------------------------------------------------
        if(sample_window[red]->channel[2] & 0x0001)
        {
            if(real_time)
                do{;} while(   sample_count * ticks_per_sample
                             > rdtsc() - time_zero
                           );
            bmp_clone_to_ezfb(&fb, &bmp);
            if(!no_blank)
                bmp_clear(&bmp, 0x00);
            frame_count++;
        }
        //-------------------------------------------------------------------
    }
    return;
}

//############################################################################
int main(int argc, char* argv[])
{
    show_tag  ();
    read_args (argc, argv);
    open_wave ();

    if(inverted)
    {
        if(setup_ezfb() && !no_ezfb_buffer)
        {
            cout << "inverted double buffer in video ram" << endl;
            ezfb_buffer_inverted();
        }
        else
        {
            cout << "inverted buffer in system ram" << endl;
            bmp_buffer_inverted();
        }
    }
    else
    {
        if(setup_ezfb() && !no_ezfb_buffer)
        {
            cout << "noninverted double buffer in video ram" << endl;
            ezfb_buffer();
        }
        else
        {
            cout << "noninverted buffer in system ram" << endl;
            bmp_buffer();
        }
    }
    //-----------------------------------------------------------------------
    cout << "sample total : " << sample_count << endl;
    cout << "frame total  : " << frame_count  << endl << endl;
    //-----------------------------------------------------------------------
    for(i = 0; i < span; i++)
        delete sample_window[i];
    //-----------------------------------------------------------------------
    //ezfb_wait_4_key();
    ezfb_release(&fb);
    return 0;
}

//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################
