//############################################################################
// ~ International Laser Display Association ~
// ~     ILDA file format driver in C++      ~
// ~             built for EZFB              ~
//
// by James Lehman
// Extra Stimulus Inc.
// james@akrobiz.com
//
// began: October 2003
//
//############################################################################
#include "ez_ilda.hpp"

//############################################################################
ez_ilda*    ez_ilda_pointer = NULL; // the one and only!

//############################################################################
ez_ilda::ez_ilda()
          : ez_ilda_frame_set       (          ),
            ez_palette_set          (          ),
            fb                      (          ),
            bmp                     (          ),
            background              (          ),
            wave_sample_rate        (48000     ), // Hz
            lit_dwell_overhang      (2         ),
            font_size_factor        (1         ),
            output_bmp_size         (1000      ), // pixels square
            viewing_angle           (          ),
            viewing_scale           (1.0       ),
            viewing_offset          (          ),
            distance_delta_max      (600       ), // points
            longest_dwell_microsec  (333       ),
            insignificant_distance  (60        ), // points
            zero_angle_tolerance    (one_degree), // radians
            rendered_arc_angle      (6.0       ), // degrees
            frames_per_second       (16        ), // 0 == every frame is shown exactly once
            still_frame_duration    (10.0      ), // seconds
            show_points             (true      ),
            show_blanking           (true      ),
            show_cursors            (false     ),
            show_palettes           (true      ),
            show_stats              (true      ),
            show_fixed_bounds       (true      ),
            show_fixed_origin       (false     ),
            show_floating_axis      (false     ),
            show_floating_bounds    (false     ),
			show_arbitrary_point    (false     ),
            show_effects_generation (true      ),
            auto_scale_dxf          (true      ),
            maintain_dxf_origin     (false     ),
            black_dwell_points      (true      ),
            auto_flatten_z          (true      ),
            auto_optimize           (true      )
{
    if(ez_ilda_pointer == NULL)
    {
        ez_ilda_pointer = this;
        ezfb_set_kb_hot();
        if(!ezfb_init_num(&fb, 0, EZFB_SAVE_SCREEN)) // as in /dev/fb0
        {
            cerr << "frame buffer failed to initialize" << endl;
            exit(12);
        }

		if(fb.Var.bits_per_pixel <= 8)
			if(!ezfb_set_bpp(&fb, 32))
			{
				cout << "32 bit frame buffer not supported" << endl;
				if(!ezfb_set_bpp(&fb, 24))
				{
					cout << "24 bit frame buffer not supported" << endl;
					if(!ezfb_set_bpp(&fb, 16))
					{
						cout << "16 bit frame buffer not supported" << endl;
						exit(666);
					}
				}
			}

        if(!ezfb_set_xy_virtual(&fb, fb.Var.xres, fb.Var.yres))
		{
			cout << "attempt to set visible and virtual display frame buffer size the same failed" << endl;
			exit(999);
		}

        ezfb_set_xy_offsets(&fb, 0, 0);
        font_size_factor = fb.Var.yres / 384;
        bmp_clone_ezfb(&fb, &bmp, 0);
        bmp_copy(&background, &bmp);

        load_effects();
#if EZ_ILDA_OUTPUT_MESSAGES
        cout << "\n\nWelcome to LaserBoy!\nmemory object initialization complete" << endl << endl;
#endif
    }
    else
    {
#if EZ_ILDA_OUTPUT_MESSAGES
        cerr << "ONLY ONE OBJECT OF class ez_ilda MAY EXIST!!!" << endl;
#endif
        exit(13);
    }
}

//############################################################################
ez_ilda::~ez_ilda()
{
    ezfb_release (&fb);
    bmp_free     (&bmp);
    bmp_free     (&background);
    ez_ilda_pointer = NULL;
}

//############################################################################
void ez_ilda::sync_rgb_and_palette()
{
    int             i                         = 0           ,
                    j                                       ,
                    super_palette_first_frame = 0           ,
                    super_palette_last_frame  = 0           ,
                    current_frame_index       = frame_index ;
    ez_palette_set  reduced_palette_set                     ;
    ez_palette      super_palette                           ;
    char            super_palette_name[7]                   ;
    static int      super_palette_number      = 0           ;
    //------------------------------------------------------------------------
    for(j = 0; j < number_of_frames(); j++)
        frame(j).sync_rgb_and_palette();
    //------------------------------------------------------------------------
    while(i < number_of_frames())
    {
        if(frame_picker(i).palette_index != EZ_TRUE_COLOR)
        {
            //--------------------------------------------------------------------
            super_palette_first_frame = i;
            super_palette = palette_picker(frame_picker(i).palette_index);
            sprintf(super_palette_name, "%06d", super_palette_number);
            super_palette.name = string("SP") + super_palette_name;
            super_palette_number++;
            //--------------------------------------------------------------------
            while(    i < number_of_frames()
                   && super_palette.unite(palette_picker(frame_picker(i).palette_index))
                 )
            i++;
            //--------------------------------------------------------------------
            reduced_palette_set.push_back(super_palette);
            super_palette_last_frame  = i - 1;
            //--------------------------------------------------------------------
            for(j = super_palette_first_frame; j <= super_palette_last_frame; j++)
            {
                frame(j).find_rgb_in_palette(super_palette);
                frame(j).palette_index = reduced_palette_set.number_of_palettes() - 1;
                frame(j).tell();
            }
            //--------------------------------------------------------------------
#if EZ_ILDA_DEBUG_MESSAGES
            cout << "super_palette_first_frame : "
                 << super_palette_first_frame
                 << " | super_palette_last_frame  : "
                 << super_palette_last_frame
                 << endl;
#endif
        }
        else
            i++;
    }
    //------------------------------------------------------------------------
    frame(current_frame_index);
    replace_palette_set(reduced_palette_set);
    palette_index = current_frame().palette_index;
    return;
}

//############################################################################
void ez_ilda::display()
{
    if(palette_index != current_frame().palette_index)
    {
        palette_index = current_frame().palette_index;
#if EZ_ILDA_MORE_OUTPUT_MESSAGES
        if(current_frame().palette_index != EZ_TRUE_COLOR)
            cout << "palette switched to " << palette_picker(palette_index).name << endl;
        else
            cout << "palette switched to TRUE_COLOR" << endl;
#endif
    }
    //------------------------------------------------------------------------
    current_frame().render_in_ezfb_bitmap();
    bmp_clone_to_ezfb(&fb, &bmp);
    return;
}

//############################################################################
void ez_ilda::load_background_bitmap(ezfb_bitmap* bmp)
{
    bmp_flip(bmp);
    bmp_into_bmp(&BACKGROUND, bmp);
    return;
}

//############################################################################
void ez_ilda::load_background_bitmap(string file)
{
    char file_name[81];
    struct ezfb_bitmap bmp;
    if(    file.size() < 4
        || file.substr(file.size() - 4) != ".bmp"
      )
        file += ".bmp";
    strcpy(file_name, (EZ_ILDA_BMP_SHARE + file).c_str());
    if(bmp_from_file(&bmp, file_name))
    {
        load_background_bitmap(&bmp);
        bmp_free(&bmp);
    }
    return;
}

//############################################################################
void ez_ilda::clear_background_bitmap()
{
    bmp_clear(&BACKGROUND, 0x00);    
    return;
}

//############################################################################
void ez_ilda::list_frame_effects()
{
    int i;
    if(frame_effect_names.size() > 0)
        for(i = 0; i < (int)frame_effect_names.size(); i++)
            cout << frame_effect_names[i] << endl;
            
#if EZ_ILDA_OUTPUT_MESSAGES
    else
        cout << "no frame effects loaded" << endl;
#endif
}

//############################################################################
void ez_ilda::list_frame_set_effects()
{
    int i;
    if(frame_set_effect_names.size() > 0)
        for(i = 0; i < (int)frame_set_effect_names.size(); i++)
            cout << frame_set_effect_names[i] << endl;
            
#if EZ_ILDA_OUTPUT_MESSAGES
    else
        cout << "no frame set effects loaded" << endl;
#endif
}

//############################################################################
void ez_ilda::call_frame_effect(string name)
{
    int i;
    for(i = 0; i < (int)frame_effect_names.size(); i++)
        if(frame_effect_names[i] == name)
            break;
            
    if(i < (int)frame_effect_names.size())
    {
#if EZ_ILDA_OUTPUT_MESSAGES
        cout << "calling " << name << " frame effect ..." << endl;
#endif
        frame_effects[i](&(current_frame()));
    }

#if EZ_ILDA_OUTPUT_MESSAGES
    else
        cout << "frame effect " << name << " not found" << endl;
#endif
}

//############################################################################
void ez_ilda::call_frame_set_effect(string name)
{
    int i;
    for(i = 0; i < (int)frame_set_effect_names.size(); i++)
        if(frame_set_effect_names[i] == name)
            break;
            
    if(i < (int)frame_set_effect_names.size())
    {
#if EZ_ILDA_OUTPUT_MESSAGES
        cout << "calling " << name << " frame set effect ..." << endl;
#endif
        frame_set_effects[i](this);
    }

#if EZ_ILDA_OUTPUT_MESSAGES
    else
        cout << "frame set effect " << name << " not found" << endl;
#endif
}

//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################






