//############################################################################
// ~ International Laser Display Association ~
// ~     ILDA file format driver in C++      ~
// ~             built for EZFB              ~
//
// by James Lehman
// Extra Stimulus Inc.
// james@akrobiz.com
//
// began: October 2003
//
//############################################################################
#include "ez_ilda.hpp"
#include "ez_wave.hpp"
    
//############################################################################
ez_ilda_frame_set::ez_ilda_frame_set(ez_ilda_point(*F)(int, int), int points_per_frame, int num_frames)
{
    for(frame_index = 0; frame_index < num_frames; frame_index++)
        push_back(ez_ilda_frame(F, points_per_frame, frame_index));
    renumber();
    frame_index = 0;
    error = EZ_ILDA_OK;
}

//############################################################################
ez_ilda_frame_set& ez_ilda_frame_set::operator += (const ez_ilda_frame& frame)
{
    if(ez_ilda_pointer->show_effects_generation)
    {
        frame.render_in_ezfb_bitmap();
        bmp_clone_to_ezfb(&(ez_ilda_pointer->fb), &(ez_ilda_pointer->bmp));
    }
    push_back(frame);
    renumber();
    return *this;
}

//############################################################################
int ez_ilda_frame_set::from_file(string file, bool append)
{
    if(    file.size() < 4
        || file.substr(file.size() - 4) != ".ild"
      )
        file += ".ild";

#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "opening " << file << " ..." << endl;
#endif

    ifstream in((EZ_ILDA_FRAMES_SHARE + file).c_str(), ios::in | ios::binary);
    if(in.is_open())
    {
        frame_index = 0;
        file_name   = file;
        error       = EZ_ILDA_OK;
        error       = from_ifstream(in, append);
        in.close();
    }
    else
    {
#if EZ_ILDA_OUTPUT_MESSAGES
        cout << file << " failed to open for input" << endl;
#endif
        error = EZ_ILDA_FILE_OPEN_FAILED;
    }
    return error;
}

//############################################################################
int ez_ilda_frame_set::from_ifstream(ifstream& in, bool append)
{
    ez_ilda_header  header                     ;
    ez_ilda_frame   frame                      ;
    ez_palette      palette                    ;
    bool            true_color_data    = false ,
                    some_data_accepted = false ;
    //------------------------------------------------------------------------
    if(!append)
    {
        clear();
        num_2D_frames = 0;
        num_3D_frames = 0;
    }
    ez_ilda_pointer->palette_index = EZ_ILDA_DEFAULT;
    //------------------------------------------------------------------------
    while(header.from_ifstream(in) == EZ_ILDA_OK)
    {   
        if(header.quantity)
        {
            switch(header.format)
            {
                //------------------------------------------------------------
                case EZ_ILDA_3D_FRAME_DATA: 
                case EZ_ILDA_2D_FRAME_DATA:
                    error |= frame.from_ifstream(in, header);
                    if(!error)
                    {
                        frame.palette_index = ez_ilda_pointer->palette_index;
                        frame.set_rgb_from_palette();
                        frame.is_2D() ? num_2D_frames++ : num_3D_frames++;
                        push_back(frame);
                        some_data_accepted = true;
                    }
                    break;
                //------------------------------------------------------------
                case EZ_ILDA_PALETTE_DATA:
                    error |= palette.from_ifstream(in, header);
                    if(!error)
                    {
                        bool built_in_palette = false;
                        for(int i = 0; i < EZ_BUILT_IN_PALETTES; i++)
                            if(header.name == ez_ilda_pointer->palette_picker(i).name)
                            {
                                ez_ilda_pointer->palette_index = i;
                                built_in_palette = true;
                                break;
                            }
                        if(!built_in_palette)
                        {
                            ez_ilda_pointer->push_back_palette(palette);
                            ez_ilda_pointer->palette_index = ez_ilda_pointer->number_of_palettes() - 1;
                        }
                        some_data_accepted = true;
                    }
                    break;
                //------------------------------------------------------------
                case EZ_ILDA_TRUE_COLOR_DATA:   
                    error |= frame.true_color_from_ifstream(in, header);
                    if(!error)
                    {
                        true_color_data = true;
                        frame.is_2D() ? num_2D_frames++ : num_3D_frames++;
                        push_back(frame);
                        some_data_accepted = true;
                    }
                    break;
                //------------------------------------------------------------
            }
            if(error)
                print_ez_ilda_error(error);
        }
#if EZ_ILDA_OUTPUT_MESSAGES
        else
            cout << "ILDA data loaded successfully" << endl;
#endif
    }
    //------------------------------------------------------------------------
    if(!in.eof() && error && some_data_accepted)
    {
#if EZ_ILDA_OUTPUT_MESSAGES
        cout << "an error occurred while reading data" << endl;
#endif
        error = EZ_ILDA_OK;
    }
    //------------------------------------------------------------------------
    if(true_color_data)
        ez_ilda_pointer->palette_index = EZ_TRUE_COLOR;
    //------------------------------------------------------------------------
    tell();
    return error;
} 

//############################################################################
int ez_ilda_frame_set::from_wave_file(string  file, bool append)
{
    if(    file.size() < 4
        || file.substr(file.size() - 4) != ".wav"
      )
        file += ".wav";

#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "opening " << file << " ..." << endl;
#endif

    ifstream in((EZ_ILDA_WAVES_SHARE + file).c_str(), ios::in | ios::binary);
    if(in.is_open())
    {
        frame_index = 0;
        file_name   = file;
        error       = EZ_ILDA_OK;
        error       = from_ifstream_wave(in, append);
        in.close();
    }
    else
    {
#if EZ_ILDA_OUTPUT_MESSAGES
        cout << file << " failed to open for input" << endl;
#endif
        error = EZ_ILDA_FILE_OPEN_FAILED;
    }
    return error;
}

//############################################################################
int ez_ilda_frame_set::from_ifstream_wave(ifstream& in, bool append)
{
    if(!append)
    {
        clear();
        num_2D_frames = 0;
        num_3D_frames = 0;
    }
    //------------------------------------------------------------------------
    int            i,
                   offset_max = -0x80000000,
                   offset_min =  0x7FFFFFFF;

    ez_ilda_frame  frame;
    ez_ilda_point  point;

    frame.format        = EZ_ILDA_2D_FRAME_DATA;
    frame.palette_index = EZ_TRUE_COLOR;
    //------------------------------------------------------------------------
    ez_wave_header header;
    if((error = header.from_ifstream(in)))
    {
#if EZ_ILDA_OUTPUT_MESSAGES
        cout << "bad wave file : " << error << endl << endl;
#endif
        return error;
    }
    header.tell();

    for(i = 0; i < header.num_channels; i++)
    {
        if(header.offset[i] > offset_max)
            offset_max = header.offset[i];
        if(header.offset[i] < offset_min)
            offset_min = header.offset[i];
    }

    int span  = (offset_max - offset_min) + 2,
        last  = span - 1,
        red   = -header.offset[2] + 1,
        green = -header.offset[3] + 1,
        blue  = -header.offset[4] + 1;

#if EZ_ILDA_DEBUG_MESSAGES
    cout << "span  : " << span   << endl
         << "red   : " << red    << endl
         << "green : " << green  << endl
         << "blue  : " << blue   << endl  << endl;
#endif

    typedef ez_wave_sample  *ez_wave_sample_pointer;
    ez_wave_sample_pointer   roll_over;
    ez_wave_sample_pointer  *sample_window;

    sample_window = new ez_wave_sample_pointer[span];
    for(i = 0; i < span; i++)
        sample_window[i] = new ez_wave_sample(header.num_channels);
    //------------------------------------------------------------------------
    while(true)
    {
        roll_over = sample_window[0];
        for(i = 0; i < last; i++)
            sample_window[i] = sample_window[i + 1];
        sample_window[last] = roll_over;

        if(sample_window[last]->from_ifstream(in))
            break;            

        point.x = sample_window[1]->channel[0];
        point.y = sample_window[1]->channel[1];
        point.z = 0;
        point.r = 255;
        point.g = 255;
        point.b = 255;
        point.k = 0;
        point.c = 0;
        //--------------------------------------------------------------------
        if(header.num_channels > 5) // rgb color
        {
            point.r = (u_char)(sample_window[red  ]->channel[2] / 128);
            point.g = (u_char)(sample_window[green]->channel[3] / 128);
            point.b = (u_char)(sample_window[blue ]->channel[4] / 128);
            if(point.r + point.g + point.b)
                point.k = 0;
            else
                point.k = EZ_ILDA_BLANKING_BIT;
            point.c = 0;

            //if(sample_window[blue]->channel[4] & 0x0001)
            //    point.set_unique();
        }
        else
        {
            point.r = (u_char)(sample_window[red  ]->channel[2] / 128);
            point.g = (u_char)(sample_window[green]->channel[2] / 128);
            point.b = (u_char)(sample_window[blue ]->channel[2] / 128);
        }
        //--------------------------------------------------------------------
        frame.push_back(point);

        if(sample_window[red]->channel[2] & 0x0001)
        {
            if(sample_window[green]->channel[3] & 0x0001)
                frame.is_unique = true;
            else
                frame.is_unique = false;

            frame.quantity = (u_short)frame.number_of_points();
            //frame.sync_rgb_and_palette();
            //frame.tell();
            push_back(frame);
            frame.clear();
        }
    }
    renumber();
    //------------------------------------------------------------------------
    for(i = 0; i < span; i++)
        delete sample_window[i];

    return EZ_ILDA_OK;
}

//############################################################################
int ez_ilda_frame_set::from_dxf_file(string file, bool append)
{
    if(    file.size() < 4
        || file.substr(file.size() - 4) != ".dxf"
      )
        file += ".dxf";

#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "opening " << file << " ..." << endl;
#endif

    ifstream in((EZ_ILDA_DXF_SHARE + file).c_str(), ios::in);
    if(in.is_open())
    {
        frame_index = 0;
        file_name   = file;
        error       = EZ_ILDA_OK;
        error       = from_ifstream_dxf(in, append);
        in.close();
    }
    else
    {
#if EZ_ILDA_OUTPUT_MESSAGES
        cout << file << " failed to open for input" << endl;
#endif
        error = EZ_ILDA_FILE_OPEN_FAILED;
    }
    return error;
}

//############################################################################
int ez_ilda_frame_set::from_dxf_directory(string dir, bool append)
{
    char             c;
    int              i;
    FILE            *pipe;
    vector<string>   file_list;    
    string           file_name,
                     command = ("ls -1 " + dir + "/*.dxf");
    //------------------------------------------------------------------------
    if(!chdir(EZ_ILDA_DXF_SHARE.c_str()))
    {
        pipe = popen(command.c_str(), "r");
        if(pipe == NULL)
        {
            cerr << "system call to {"
                 << command
                 << "} failed."
                 << endl;
            ez_ilda_terminate(444); // must exit process
        }
        //------------------------------------------------------------------------
        while(1)
        {
            c = fgetc(pipe);
            if(isgraph(c))
                file_name += c;
            else
            {
                file_list.push_back(file_name);
                file_name.clear();
            }
            if(c == EOF)
                break;
        }
        //------------------------------------------------------------------------
        pclose(pipe);

        chdir("../");

        if(!append)
        {
            clear();
            num_2D_frames = 0;
            num_3D_frames = 0;
        }
        //------------------------------------------------------------------------
        for(i = 0; i < (int)file_list.size(); i++)
            error |= from_dxf_file(file_list[i], true);
        //------------------------------------------------------------------------
        renumber();
    }
    else
    {
#if EZ_ILDA_OUTPUT_MESSAGES
        cout << dir << " failed to open for input" << endl;
#endif
        error = EZ_ILDA_FILE_OPEN_FAILED;
    }
    return error;
}

//############################################################################
int ez_ilda_frame_set::from_ifstream_dxf(ifstream& in, bool append)
{
    ez_ilda_header  header  ;
    ez_ilda_frame   frame   ;
    ez_palette      palette ;
    //------------------------------------------------------------------------
    if(!append)
    {
        clear();
        num_2D_frames = 0;
        num_3D_frames = 0;
    }
    //------------------------------------------------------------------------
    frame.format        = EZ_ILDA_3D_FRAME_DATA;
    frame.palette_index = EZ_DXF;
    //------------------------------------------------------------------------
    if(    (error = frame.from_ifstream_dxf(in, header))
        == EZ_ILDA_OK
      )
    {
        frame.quantity = (u_short)frame.number_of_points();
        frame.sync_rgb_and_palette();
        frame.tell();
        push_back(frame);
        ez_ilda_pointer->palette_index = EZ_DXF;
        renumber();
    }
    //------------------------------------------------------------------------
    tell();
    return error;
}

//############################################################################
int ez_ilda_frame_set::from_shp_file(string file)
{
    if(    file.size() < 4
        || file.substr(file.size() - 4) != ".shp"
      )
        file += ".shp";

#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "opening " << file << " ..." << endl;
#endif

    ifstream in((EZ_ILDA_DXF_SHARE + file).c_str(), ios::in);
    if(in.is_open())
    {
        frame_index = 0;
        file_name   = file;
        error       = EZ_ILDA_OK;

        error       = from_ifstream_shp(in);
        in.close();
    }
    else
    {
#if EZ_ILDA_OUTPUT_MESSAGES
        cout << file << " failed to open for input" << endl;
#endif
        error = EZ_ILDA_FILE_OPEN_FAILED;
    }
    return error;
}

//############################################################################
int ez_ilda_frame_set::from_ifstream_shp(ifstream& in)
{
    ez_ilda_header  header  ;
    ez_ilda_frame   frame   ;
    ez_palette      palette ;
    //------------------------------------------------------------------------
    clear();
    num_2D_frames = 0;
    num_3D_frames = 0;
    //------------------------------------------------------------------------
    frame.format        = EZ_ILDA_2D_FRAME_DATA;
    frame.palette_index = EZ_DXF;
    //------------------------------------------------------------------------
    while(    (error = frame.from_ifstream_shp(in, header))
           == EZ_ILDA_OK
      )
    {
        frame.quantity = (u_short)frame.number_of_points();
        frame.sync_rgb_and_palette();
        frame.tell();
        push_back(frame);
    }
    //------------------------------------------------------------------------
    ez_ilda_pointer->palette_index = EZ_DXF;
    renumber();
    tell();
    return error;
}

//############################################################################
void ez_ilda_frame_set::flatten_z()
{
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "flattening Z ..." << endl;
#endif
    for(int i = 0; i < number_of_frames(); i++)
        at(i).flatten_z();
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "flatten Z DONE!" << endl;
#endif
    return;
}

//############################################################################
void ez_ilda_frame_set::z_order_points(unsigned short span)
{
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "Z ordering points ..." << endl;
#endif
    for(int i = 0; i < number_of_frames(); i++)
        at(i).z_order_points(span);
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "Z order DONE!" << endl;
#endif
    return;
}

//############################################################################
void ez_ilda_frame_set::minimize_blank_points()
{
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "minimizing points ..." << endl;
#endif
    for(int i = 0; i < number_of_frames(); i++)
        at(i).minimize_blank_points();
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "minimize points DONE!" << endl;
#endif
    return;
}

//############################################################################
void ez_ilda_frame_set::remove_dwell_points()
{
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "removing dwell points ..." << endl;
#endif
    for(int i = 0; i < number_of_frames(); i++)
        at(i).remove_dwell_points();
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "removing dwell points DONE!" << endl;
#endif
    return;
}

//############################################################################
void ez_ilda_frame_set::minimize_vectors()
{
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "minimizing vectors ..." << endl;
#endif
    for(int i = 0; i < number_of_frames(); i++)
        at(i).minimize_vectors();
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "minimize vectors DONE!" << endl;
#endif
    return;
}

//############################################################################
void ez_ilda_frame_set::minimize_points()
{
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "minimizing points ..." << endl;
#endif
    for(int i = 0; i < number_of_frames(); i++)
        at(i).minimize_points();
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "minimize points DONE!" << endl;
#endif
    return;
}

//############################################################################
void ez_ilda_frame_set::explode_current_frame()
{
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "exploding current frame ..." << endl;
#endif
    int                i;
    ez_ilda_frame_set  results,
                       exploded = current_frame().find_segments();

    for(i = 0; i < frame_index; i++)
        results += at(i);

    for(i = 0; i < exploded.number_of_frames(); i++)
        results += exploded[i];

    for(i = frame_index + 1; i < number_of_frames(); i++)
        results += at(i);

    *this = results;
    renumber();
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "exploding current frame DONE!" << endl;
#endif
    return;
}

//############################################################################
void ez_ilda_frame_set::collapse_selected_frames()
{
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "collapsing selected frames ..." << endl;
#endif
    ez_ilda_frame     collapsed;
    int               first_selected_frame_index = -1;

    for(int i = 0; i < number_of_frames(); i++)
    {
        if(at(i).is_selected)
        {
            if(first_selected_frame_index == -1)
                first_selected_frame_index = i;
            collapsed += at(i);
        }
    }
    if(first_selected_frame_index != -1)
    {
        frame(first_selected_frame_index) = collapsed;
        frame(first_selected_frame_index).is_selected = false;
        delete_selected_frames();
        renumber();
    }
#if EZ_ILDA_OUTPUT_MESSAGES
    else
    {
        cout << "no frames have been selected" << endl;
    }
    cout << "collapsing selected frames DONE!" << endl;
#endif
    return;
}

//############################################################################
void ez_ilda_frame_set::fracture_segments()
{
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "fracturing segments ..." << endl;
#endif
    for(int i = 0; i < number_of_frames(); i++)
        at(i).fracture_segments();
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "fracturing segments DONE!" << endl;
#endif
    return;
}

//############################################################################
void ez_ilda_frame_set::bond_segments()
{
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "bonding segments ..." << endl;
#endif
    for(int i = 0; i < number_of_frames(); i++)
        at(i).bond_segments();
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "bonding segments DONE!" << endl;
#endif
    return;
}

//############################################################################
void ez_ilda_frame_set::reorder_segments()
{
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "reordering segments ..." << endl;
#endif
    for(int i = 0; i < number_of_frames(); i++)
        at(i).reorder_segments();
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "reordering segments DONE!" << endl;
#endif
    return;
}

//############################################################################
void ez_ilda_frame_set::add_dwell_points()
{
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "adding dwell points ..." << endl;
#endif
    for(int i = 0; i < number_of_frames(); i++)
        at(i).add_dwell_points();
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "adding dwell points DONE!" << endl;
#endif
    return;
}

//############################################################################
void ez_ilda_frame_set::add_span_points()
{
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "adding span points ..." << endl;
#endif
    for(int i = 0; i < number_of_frames(); i++)
        at(i).add_span_points();
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "adding span points DONE!" << endl;
#endif
    return;
}

//############################################################################
void ez_ilda_frame_set::delete_redundant_frames()
{
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "deleting redundant frames ..." << endl;
#endif
    ez_ilda_frame_set reduced;
    for(int i = 0; i < number_of_frames(); i++)
        if(at(i).is_unique)
            reduced.push_back(at(i));
    reduced.renumber();
    *this = reduced;
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "delete redundant frames DONE!" << endl;
#endif
    return;
}

//############################################################################
void ez_ilda_frame_set::optimize_points()
{
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "optimizing points ..." << endl;
#endif
    if(number_of_frames())
    {
        int i,
            dwell_point_count = 0,
            dwell_point_total = 0,
            blank_dwell       = ez_ilda_pointer->wave_sample_rate / 220;
        //--------------------------------------------------------------------
        for(i = 0; i < number_of_frames(); i++)
            at(i).optimize_points();
        //--------------------------------------------------------------------
        intro.clear();
        for(i = 0; i < blank_dwell; i++)
            intro.push_back(ez_ilda_point()); // origin
        //--------------------------------------------------------------------
        intro += ez_ilda_segment(intro[0], front()[0]); // from origin to first point in first frame
        //--------------------------------------------------------------------
        for(i = 0; i < blank_dwell; i++)
            intro.push_back(front()[0]);
        //--------------------------------------------------------------------
        for(i = 0; i < (number_of_frames() - 1); i++)
        {
            at(i).transition_to_next_frame.clear();
            dwell_point_total = dwell_points(delta_angle(at(i)[at(i).number_of_points() - 2] | at(i + 1).front(),
                                                         at(i)[at(i).number_of_points() - 2] | at(i).back(),
                                                         at(i).back()                        | at(i + 1).front()
                                                        )
                                            );
            if(dwell_point_total)
                for(dwell_point_count = 0; dwell_point_count < dwell_point_total; dwell_point_count++)
                    at(i).transition_to_next_frame.push_back(at(i).back());
            //----------------------------------------------------------------            
            at(i).transition_to_next_frame += ez_ilda_segment(at(i).back(), at(i + 1)[0]);
        }
        //--------------------------------------------------------------------
        back().transition_to_next_frame = ez_ilda_segment(back().back(), intro[0]);
        for(i = 0; i < blank_dwell; i++)
            back().transition_to_next_frame.push_back(intro[0]);
        //--------------------------------------------------------------------
    }
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "optimize points DONE!" << endl;
#endif
    return;
}

//############################################################################
void ez_ilda_frame_set::rainbow_recolor(int effect, float factor)
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).rainbow_recolor(effect, factor);
    return;
}

//############################################################################
void ez_ilda_frame_set::quarter_turn(int plane, int turns)
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).quarter_turn(plane, turns);
    return;
}

//############################################################################
int ez_ilda_frame_set::scale_around_origin(ez_3D_float factor)
{
    int all_in_bounds = true;
    for(int i = 0; i < number_of_frames(); i++)
        all_in_bounds &= at(i).scale_around_origin(factor);
    return all_in_bounds;
}

//############################################################################
int ez_ilda_frame_set::scale_around_origin()
{
    int all_in_bounds = true;
    for(int i = 0; i < number_of_frames(); i++)
        all_in_bounds &= at(i).scale_around_origin();
    return all_in_bounds;
}

//############################################################################
int ez_ilda_frame_set::rotate_around_origin(ez_3D_float angle)
{
    int all_in_bounds = true;
    for(int i = 0; i < number_of_frames(); i++)
        all_in_bounds &= at(i).rotate_around_origin(angle);
    return all_in_bounds;
}

//############################################################################
int ez_ilda_frame_set::rotate_around_origin()
{
    int all_in_bounds = true;
    for(int i = 0; i < number_of_frames(); i++)
        all_in_bounds &= at(i).rotate_around_origin();
    return all_in_bounds;
}

//############################################################################
int ez_ilda_frame_set::move(ez_3D_float offset)
{
    int all_in_bounds = true;
    for(int i = 0; i < number_of_frames(); i++)
        all_in_bounds &= at(i).move(offset);
    return all_in_bounds;
}

//############################################################################
int ez_ilda_frame_set::move()
{
    int all_in_bounds = true;
    for(int i = 0; i < number_of_frames(); i++)
        all_in_bounds &= at(i).move();
    return all_in_bounds;
}

//############################################################################
void ez_ilda_frame_set::toggle_frame_select_start()
{
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "frame select start was ";        
    if(frame_select_start == -1)
        cout << "unset";
    else
        cout << frame_select_start;
    cout << endl;
#endif
    //------------------------------------------------------------------------
    if(frame_select_start == -1) // was unset
    {
        frame_select_start = frame_index; // set
        if(frame_select_end != -1)
        {
            for(int i = frame_select_start;
                    i != frame_select_end;
                    (frame_select_start <= frame_select_end) ? (i++) : (i--)
               )
                at(i).is_selected = true;
            at(frame_select_end).is_selected = true;
        }
    }
    else // was set
    {
        if(frame_select_end != -1) // was set
        {
            for(int i = frame_select_start;
                    i != frame_select_end;
                    (frame_select_start <= frame_select_end) ? (i++) : (i--)
               )
                at(i).is_selected = false;
            at(frame_select_end).is_selected = false;
        }
        frame_select_start = -1; // unset
    }
    //------------------------------------------------------------------------
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "frame select start is ";        
    if(frame_select_start == -1)
        cout << "unset";
    else
        cout << frame_select_start;
    cout << endl;
#endif
    return;
}

//############################################################################
void ez_ilda_frame_set::toggle_frame_select_end()
{
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "frame select end was ";
    if(frame_select_end == -1)
        cout << "unset";
    else
        cout << frame_select_end;
    cout << endl;
#endif    
    //------------------------------------------------------------------------
    if(frame_select_end == -1) // was unset
    {
        frame_select_end = frame_index; // set
        if(frame_select_start != -1)
        {
            for(int i = frame_select_start;
                    i != frame_select_end;
                    (frame_select_start <= frame_select_end) ? (i++) : (i--)
               )
                at(i).is_selected = true;
            at(frame_select_end).is_selected = true;
        }
    }
    else // was set
    {
        if(frame_select_start != -1) // was set
        {
            for(int i = frame_select_start;
                    i != frame_select_end;
                    (frame_select_start <= frame_select_end) ? (i++) : (i--)
               )
                at(i).is_selected = false;
            at(frame_select_end).is_selected = false;
        }
        frame_select_end = -1; // unset
    }
    //------------------------------------------------------------------------
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "frame select end is ";        
    if(frame_select_end == -1)
        cout << "unset";
    else
        cout << frame_select_end;
    cout << endl;
#endif
    return;
}

//############################################################################
void ez_ilda_frame_set::delete_selected_frames()
{
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "deleting selected frames ..." << endl;
#endif
    ez_ilda_frame_set reduced;
    for(int i = 0; i < number_of_frames(); i++)
    {
        if(!at(i).is_selected)
            reduced.push_back(at(i));
    }
    reduced.renumber();

    if(    frame_select_start != -1
        && frame_select_end   != -1
      )
    {
        frame_index = (frame_select_start - 1);
        frame_select_start = -1;
        frame_select_end   = -1;
    }

    if(frame_index < 0)
        frame_index = 0;

    if(frame_index > reduced.number_of_frames())
        frame_index = reduced.number_of_frames() - 1;

    *this = reduced;
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "deleting selected frames DONE!" << endl;
#endif
    return;
}

//############################################################################
void ez_ilda_frame_set::trim_to_selected_frames()
{
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "trimming to selected frames ..." << endl;
#endif
    ez_ilda_frame_set reduced;
    for(int i = 0; i < number_of_frames(); i++)
    {
        if(at(i).is_selected)
            reduced.push_back(at(i));
    }
    reduced.renumber();
    if(    frame_select_start != -1
        && frame_select_end   != -1
      )
    {
        frame_select_start = -1;
        frame_select_end   = -1;
    }
    frame_index = 0;
    *this = reduced;
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "trim to selected frames DONE!" << endl;
#endif
    return;
}

//############################################################################
void ez_ilda_frame_set::bit_reduce_to_palette()
{
#if EZ_ILDA_DEBUG_MESSAGES
    cout << "reducing to palette ..." << endl;
#endif
    for(int i = 0; i < number_of_frames(); i++)
        at(i).bit_reduce_to_palette();
#if EZ_ILDA_DEBUG_MESSAGES
    cout << "reduced to palette DONE!" << endl;
#endif
    return;
}

//############################################################################
void ez_ilda_frame_set::promote_to_true_color()
{
#if EZ_ILDA_DEBUG_MESSAGES
    cout << "promoting to true color ..." << endl;
#endif
    for(int i = 0; i < number_of_frames(); i++)
        at(i).promote_to_true_color();
#if EZ_ILDA_DEBUG_MESSAGES
    cout << "promoting to true color DONE!" << endl;
#endif
    return;
}

//############################################################################
void ez_ilda_frame_set::best_match_palette(int index)
{
#if EZ_ILDA_DEBUG_MESSAGES
    cout << "matching to palette " << ez_ilda_pointer->palette_picker(index).name << " ..." << endl;
#endif
    for(int i = 0; i < number_of_frames(); i++)
        at(i).best_match_palette(index);
#if EZ_ILDA_DEBUG_MESSAGES
    cout << "matched to palette DONE!" << endl;
#endif
    return;
}

//############################################################################
void ez_ilda_frame_set::best_match_target_palette()
{
#if EZ_ILDA_DEBUG_MESSAGES
    cout << "matching target palette ..." << endl;
#endif
    for(int i = 0; i < number_of_frames(); i++)
        at(i).best_match_target_palette();
#if EZ_ILDA_DEBUG_MESSAGES
    cout << "matched target palette DONE!" << endl;
#endif
    return;
}

//############################################################################
void ez_ilda_frame_set::accept_palette(int index)
{
#if EZ_ILDA_DEBUG_MESSAGES
    cout << "accepting palette  " << ez_ilda_pointer->palette_picker(index).name << " ..." << endl;
#endif
    for(int i = 0; i < number_of_frames(); i++)
        at(i).accept_palette(index);
#if EZ_ILDA_DEBUG_MESSAGES
    cout << "accept palette DONE!" << endl;
#endif
    return;
}

//############################################################################
void ez_ilda_frame_set::accept_target_palette()
{
#if EZ_ILDA_DEBUG_MESSAGES
    cout << "accept target palette ..." << endl;
#endif
    for(int i = 0; i < number_of_frames(); i++)
        at(i).accept_target_palette();
#if EZ_ILDA_DEBUG_MESSAGES
    cout << "accept target palette DONE!" << endl;
#endif
    return;
}

//############################################################################
void ez_ilda_frame_set::color_over_ezfb_bitmap(ezfb_bitmap* bmp)
{
#if EZ_ILDA_DEBUG_MESSAGES
    cout << "coloring over ezfb bitmap ..." << endl;
#endif
    for(int i = 0; i < number_of_frames(); i++)
        at(i).color_over_ezfb_bitmap(bmp);
#if EZ_ILDA_DEBUG_MESSAGES
    cout << "color over ezfb bitmap DONE!" << endl;
#endif
    return;
}

//############################################################################
void ez_ilda_frame_set::color_over_ezfb_bitmap(string file)
{
    char file_name[81];
    struct ezfb_bitmap bmp;
    if(    file.size() < 4
        || file.substr(file.size() - 4) != ".bmp"
      )
        file += ".bmp";
    strcpy(file_name, (EZ_ILDA_BMP_SHARE + file).c_str());
    if(bmp_from_file(&bmp, file_name))
    {
        color_over_ezfb_bitmap(&bmp);
        bmp_free(&bmp);
    }
    return;
}

//############################################################################
void ez_ilda_frame_set::flip(int plane)
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).flip(plane);
    return;
}

//############################################################################
void ez_ilda_frame_set::renumber()
{
    int   i,
          local_palette_index  = EZ_ILDA_DEFAULT,
          added_palettes       = 0,
          frame_id             = 0;
    bool  in_true_color        = false;
    char  number[7];
    //------------------------------------------------------------------------
    for(int i = 0; i < number_of_frames(); i++)
    {
        if((at(i).quantity = at(i).number_of_points()) > 1)
        {
            if(at(i).palette_index == EZ_TRUE_COLOR)
            {
                if(!in_true_color)
                {
                    frame_id++;
                    added_palettes++; // add EZFB_332
                    local_palette_index = EZ_EZFB_332;
                    in_true_color = true;
                }
                frame_id++;
                added_palettes++; // add true_color
            }
            else if(at(i).palette_index != local_palette_index)
            {
                frame_id++;
                added_palettes++;
                local_palette_index = at(i).palette_index;
                in_true_color = false;
            }
            sprintf(number, "%06d", frame_id);
            at(i).identity = frame_id;
            at(i).name     = "ES" + string(number);
            frame_id++;
        }
        else
            delete_frame(i);
    }
    //------------------------------------------------------------------------
    for(i = 0; i < number_of_frames(); i++)
        at(i).total = (u_short)(number_of_frames() + added_palettes);
    if(frame_index >= number_of_frames())
        frame_index = number_of_frames() - 1;
    return;
}

//############################################################################
ez_ilda_frame_set& ez_ilda_frame_set::reverse()
{
    ez_ilda_frame_set reversed;
    for(int i = 1; i <= number_of_frames(); i++)
        reversed.push_back(at(number_of_frames() - i));
    clear();
    insert(begin(), reversed.begin(), reversed.end());
    renumber();
    tell();
    return *this;
}

//############################################################################
ez_ilda_frame_set& ez_ilda_frame_set::frame_reverse()
{
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "reversing point order in frames ..." << endl;
#endif
    for(int i = 1; i <= number_of_frames(); i++)
        at(i).reverse();        
    tell();
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "reversing point order in frames DONE!" << endl;
#endif
    return *this;
}

//############################################################################
ez_ilda_frame_set& ez_ilda_frame_set::superimpose_from_file(string file)
{
    ez_ilda_frame_set  overlay,
                       sum    ;

    int                i;
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "superimposing " << file << " ..." << endl;
#endif
    overlay.from_file(file);

    if(overlay.number_of_frames() > 1 && number_of_frames() > 1)
    {
        if(number_of_frames() > overlay.number_of_frames())
        {
            for(i = 0; i < overlay.number_of_frames(); i++)
                sum += (frame(i) + overlay.frame(i));

            for(i = overlay.number_of_frames(); i < number_of_frames(); i++)
                sum += frame(i);
        }
        else if(number_of_frames() < overlay.number_of_frames())
        {
            for(i = 0; i < number_of_frames(); i++)
                sum += (frame(i) + overlay.frame(i));

            for(i = number_of_frames(); i < overlay.number_of_frames(); i++)
                sum += overlay.frame(i);
        }
        else
        {
            for(i = 0; i < number_of_frames(); i++)
                sum += (frame(i) + overlay.frame(i));
        }
    }
    else if(overlay.number_of_frames() == 1)
    {
        for(i = 0; i < number_of_frames(); i++)
            sum += (frame(i) + overlay.frame(0));
    }
    else if(number_of_frames() == 1)
    {
        for(i = 0; i < overlay.number_of_frames(); i++)
            sum += (frame(0) + overlay.frame(i));
    }
    *this = sum;
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "superimposing " << file << " DONE!" << endl;
#endif
    return *this;
}

//############################################################################
void ez_ilda_frame_set::save_as_ilda(string file)
{
    if(    file.size() < 4
        || file.substr(file.size() - 4) != ".ild"
      )
        file += ".ild";

#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "saving " << file << " ..." << endl;
#endif
    ofstream out((EZ_ILDA_FRAMES_SHARE + file).c_str(), ios::out | ios::binary);
    if(out.is_open())
    {
        to_ofstream(out);
        file_name = file;
        out.close();
#if EZ_ILDA_OUTPUT_MESSAGES
        cout << file_name << " saved" << endl;
#endif
    }
#if EZ_ILDA_OUTPUT_MESSAGES
    else
        cout << file << " failed to open for output" << endl;
#endif
    return;
}

//############################################################################
void ez_ilda_frame_set::to_ofstream(ofstream& out)
{
    int             file_section_number = 0,
                    local_palette_index = EZ_ILDA_DEFAULT;
    bool            in_true_color       = false;
    ez_ilda_header  header;
    //------------------------------------------------------------------------
    renumber();
    header.format  = EZ_ILDA_PALETTE_DATA;
    header.total   = front().total;    
    //------------------------------------------------------------------------    
    for(int i = 0; i < number_of_frames(); i++)
    {
        if(at(i).palette_index == EZ_TRUE_COLOR)
        {
            if(!in_true_color)
            {
                header.name         = ez_ilda_pointer->palette_picker(EZ_EZFB_332).name;
                header.quantity     = ez_ilda_pointer->palette_picker(EZ_EZFB_332).number_of_colors();
                header.identity     = file_section_number;
                ez_ilda_pointer->palette_picker(EZ_EZFB_332).to_ofstream(out, header);
                local_palette_index = EZ_EZFB_332;
                file_section_number++;
                in_true_color = true;
            }
            file_section_number++;
        }
        else if(at(i).palette_index != local_palette_index)
        {
            header.name         = ez_ilda_pointer->palette_picker(at(i).palette_index).name;
            header.quantity     = ez_ilda_pointer->palette_picker(at(i).palette_index).number_of_colors();
            header.identity     = file_section_number;
            ez_ilda_pointer->palette_picker(at(i).palette_index).to_ofstream(out, header);
            local_palette_index = at(i).palette_index;
            file_section_number++;
            in_true_color = false;
        }
        at(i).is_2D(); // checks for z and sets format accordingly;
        at(i).to_ofstream(out);
        file_section_number++;
    }
    header.format   = EZ_ILDA_3D_FRAME_DATA ;
    header.name     = "ILDA_End"            ;
    header.quantity = 0                     ;
    header.identity = file_section_number   ;
    header.to_ofstream(out)                 ;
    return;
}

//############################################################################
void ez_ilda_frame_set::print_ez_ilda_error(int error) const
{
#if EZ_ILDA_OUTPUT_MESSAGES
    if(!error)
        cout << "    EZ_ILDA_OK" << endl;
    else
    {
        if(error & EZ_ILDA_BAD_HEADER           )  cout << "    EZ_ILDA_BAD_HEADER"           << endl;
        if(error & EZ_ILDA_EOF_ERROR            )  cout << "    EZ_ILDA_EOF_ERROR"            << endl;
        if(error & EZ_ILDA_FILE_OPEN_FAILED     )  cout << "    EZ_ILDA_FILE_OPEN_FAILED"     << endl;
        if(error & EZ_ILDA_UNKNOWN_FORMAT       )  cout << "    EZ_ILDA_UNKNOWN_FORMAT"       << endl;
        if(error & EZ_ILDA_INCOMPATIBLE_FORMATS )  cout << "    EZ_ILDA_INCOMPATIBLE_FORMATS" << endl;
    }
#endif
    return;
}

//############################################################################
void ez_ilda_frame_set::tell() const
{
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "\n    total frames    : "
         << number_of_frames()
         << "\n    total 2D frames : "
         << num_2D_frames
         << "\n    total 3D frames : "
         << num_3D_frames
         << "\n    total palettes  : "
         << ez_ilda_pointer->number_of_palettes()
         << "\n    new   palettes  : "
         << ez_ilda_pointer->number_of_palettes() - EZ_BUILT_IN_PALETTES
         << "\n    total points    : "
         << total_points()
         << endl << endl;
#endif
    return;
}

//############################################################################
int ez_ilda_frame_set::total_wave_samples() const
{
    int   i,
          samples = intro.number_of_points();

    for(i = 0; i < number_of_frames(); i++)
    {
        samples += (   (   (ez_ilda_pointer->frames_per_second > 0)
                         ? (int)ceil(   (    ez_ilda_pointer->wave_sample_rate 
                                           / at(i).number_of_points()
                                         )
                                       / ez_ilda_pointer->frames_per_second
                                    )
                         : 1
                       )
                     * at(i).number_of_points()
                     + at(i).transition_to_next_frame.number_of_points()
                   );
    }
#if EZ_ILDA_DEBUG_MESSAGES
    cout << "total_wave_samples : " << samples << endl << endl;
#endif
    return samples;
}

//############################################################################
void ez_ilda_frame_set::save_as_wave(string file) const
{
    if(    file.size() < 4
        || file.substr(file.size() - 4) != ".wav"
      )
        file += ".wav";

#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "saving " << file << " ..." << endl;
#endif
    ofstream out((EZ_ILDA_WAVES_SHARE + file).c_str(), ios::out | ios::binary);
    if(out.is_open())
    {
        if(number_of_frames() > 1)
            to_ofstream_wave(out);
        else
        {
            ez_ilda_frame copy(at(0));

            copy.rotate_around_origin();

            if(ez_ilda_pointer->auto_flatten_z)
                copy.flatten_z();

            if(ez_ilda_pointer->auto_optimize)
                copy.optimize_points();

            copy.save_as_wave(file);
        }

        out.close();
#if EZ_ILDA_OUTPUT_MESSAGES
        cout << file << " saved" << endl;
#endif
    }
#if EZ_ILDA_OUTPUT_MESSAGES
    else
        cout << file << " failed to open for output" << endl;
#endif
    return;
}

//############################################################################
void ez_ilda_frame_set::to_ofstream_wave(ofstream& out) const
{
    ez_ilda_frame_set copy(*this);

    for(int i = 0; i < copy.number_of_frames(); i++)
        copy[i].rotate_around_origin();

    if(ez_ilda_pointer->auto_flatten_z)
        copy.flatten_z();

    if(ez_ilda_pointer->auto_optimize)
        copy.optimize_points();

    ez_wave_header  header(copy.total_wave_samples(), ez_ilda_pointer->wave_sample_rate);

    header.to_ofstream(out);

    if(ez_ilda_pointer->auto_optimize)
        copy.intro.to_ofstream_wave(out);

    for(int i = 0; i < copy.number_of_frames(); i++)
        copy.at(i).to_ofstream_wave(out);

    return;
}

//############################################################################
void ez_ilda_frame_set::save_as_bmp_directory(string dir) const
{
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "saving " << dir << " ..." << endl;
#endif
    string command_string;
    char name[8];
    command_string  = "mkdir " + EZ_ILDA_BMP_SHARE + dir;
    call_command(command_string, ez_ilda_terminate, false);
    for(int i = 0; i < number_of_frames(); i++)
    {
        sprintf(name, "%06d", i);
        at(i).save_as_bmp(string(dir + '/' + name));
    }
    return;
}

//############################################################################
void ez_ilda_frame_set::save_as_dxf_directory(string dir) const
{
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "saving " << dir << " ..." << endl;
#endif
    string command_string;
    char name[8];
    command_string  = "mkdir " + EZ_ILDA_DXF_SHARE + dir;
    call_command(command_string, ez_ilda_terminate, false);
    for(int i = 0; i < number_of_frames(); i++)
    {
        sprintf(name, "%06d", i);
        at(i).save_as_dxf(string(dir + "/" + name));
    }
    return;
}

//############################################################################
void ez_ilda_frame_set::save_as_code(string object_name)
{
#if EZ_ILDA_OUTPUT_MESSAGES
    cout << "saving " << object_name << ".cpp ..." << endl;
#endif
    ofstream out((EZ_ILDA_CPP_SHARE + object_name + ".cpp").c_str(), ios::out);
    if(out.is_open())
    {
        out << "//############################################################################\n"
               "// ~  This file is the product of LaserBoy!  ~\n"
               "// ~ International Laser Display Association ~\n"
               "// ~    ez_ilda file format driver in C++    ~\n"
               "//\n"
               "// by James Lehman\n"
               "// james@akrobiz.com\n"
               "//\n"
               "//############################################################################\n\n\n\n"
               "//############################################################################\n"
            << "ez_ilda_frame_set ";
        out << object_name;
        out << "()\n"
               "{\n"
               "    ez_ilda_frame     frame;\n"
               "    ez_ilda_frame_set frames;\n"
               "    int               frame_index;\n\n"
               "    for(frame_index = 0; frame_index < ";                   

        out << number_of_frames();

        out << "; frame_index++)\n"
               "    {\n";

        for(int i = 0; i < number_of_frames(); i++)
        {

            at(i).to_ofstream_code(out, "        ");
            
            out << "        frames.push_back(frame);\n"
                   "        frame.clear();\n\n";
        }

        out << "    }\n"
               "    return frames;\n"
               "}\n\n"
               "//############################################################################\n"
               "//////////////////////////////////////////////////////////////////////////////\n"
               "//############################################################################\n\n\n";

        out.close();
#if EZ_ILDA_OUTPUT_MESSAGES
        cout << object_name << ".cpp saved" << endl;
#endif
    }
#if EZ_ILDA_OUTPUT_MESSAGES
    else
        cout << object_name << ".cpp failed to open for output" << endl;
#endif
    return;
}


//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################









